import json
from copy import deepcopy
from unittest import IsolatedAsyncioTestCase
from unittest.mock import MagicMock

import anoncreds
import pytest
from anoncreds import (
    AnoncredsError,
    AnoncredsErrorCode,
    Credential,
    CredentialDefinition,
    CredentialRequest,
    CredentialRevocationState,
    Presentation,
    PresentCredentials,
    RevocationRegistryDefinition,
    Schema,
    W3cCredential,
    W3cPresentation,
)
from aries_askar import AskarError, AskarErrorCode
from pyld import jsonld
from pyld.jsonld import JsonLdProcessor

from ...anoncreds.tests.mock_objects import (
    CRED_DEFS,
    CREDS_W3C_METADATA,
    MOCK_CRED,
    MOCK_CRED_DEF,
    MOCK_PRES,
    MOCK_PRES_REQ,
    MOCK_W3C_CRED,
    MOCK_W3C_PRES_REQ,
    MOCK_W3CPRES,
    SCHEMAS,
)
from ...askar.profile_anon import AskarAnonCredsProfile, AskarAnonCredsProfileSession
from ...tests import mock
from ...utils.testing import create_test_profile
from ...vc.ld_proofs.document_loader import DocumentLoader
from ...wallet.error import WalletNotFoundError
from .. import holder as test_module
from ..holder import CATEGORY_CREDENTIAL, AnonCredsHolder, AnonCredsHolderError
from ..models.credential_definition import CredDef, CredDefValue, CredDefValuePrimary
from ..models.revocation import GetRevListResult, RevList
from ..registry import AnonCredsRegistry


class MockCredReceived:
    def __init__(self, bad_schema=False, bad_cred_def=False):
        self.schema_id = "Sc886XPwD1gDcHwmmLDeR2:2:degree schema:45.101.94"
        self.cred_def_id = "Sc886XPwD1gDcHwmmLDeR2:3:CL:229975:faber.agent.degree_schema"

    schema_id = "Sc886XPwD1gDcHwmmLDeR2:2:degree schema:45.101.94"
    cred_def_id = "Sc886XPwD1gDcHwmmLDeR2:3:CL:229975:faber.agent.degree_schema"
    rev_reg_id = None

    def to_json_buffer(self):
        return b"credential"

    def to_dict(self):
        return MOCK_CRED


class MockCredReceivedW3C:
    def __init__(self):
        self.schema_id = "Sc886XPwD1gDcHwmmLDeR2:2:degree schema:45.101.94"
        self.cred_def_id = "Sc886XPwD1gDcHwmmLDeR2:3:CL:229975:faber.agent.degree_schema"

    def to_json_buffer(self):
        return b"credential"

    def to_dict(self):
        return MOCK_W3C_CRED


class MockCredential:
    def __init__(self):
        self.rev_reg_id = "rev-reg-id"
        self.rev_reg_index = 0

    cred = mock.AsyncMock(auto_spec=Credential)

    def to_dict(self):
        return MOCK_CRED

    def process(self, *args, **kwargs):
        return MockCredReceived()


class MockW3Credential:
    cred = mock.AsyncMock(auto_spec=W3cCredential)

    def to_dict(self):
        return MOCK_W3C_CRED

    def process(self, *args, **kwargs):
        return MockCredReceivedW3C()


class MockMasterSecret:
    value = b"master-secret"


class MockCredEntry:
    def __init__(self, rev_reg=False) -> None:
        mock_cred = deepcopy(MOCK_CRED)
        if rev_reg:
            mock_cred["rev_reg_id"] = "rev-reg-id"

        mock_cred["rev_reg_index"] = 1
        self.name = "name"
        self.value = mock_cred
        self.raw_value = mock_cred

    def decode(self):
        return MOCK_CRED


class MockCredScan:
    value = [MockCredEntry()]
    name = "name"

    def __aiter__(self):
        return self

    async def __anext__(self):
        if self.value:
            return self.value.pop()
        raise StopAsyncIteration


class MockMimeTypeRecord:
    value_json = {"mime-type": "mime-type"}


@pytest.mark.anoncreds
class TestAnonCredsHolder(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile(
            settings={"wallet.type": "askar-anoncreds"}
        )
        self.holder = test_module.AnonCredsHolder(self.profile)

    async def test_init(self):
        assert isinstance(self.holder, AnonCredsHolder)
        assert isinstance(self.holder.profile, AskarAnonCredsProfile)

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_get_master_secret(self, mock_session_handle):
        mock_session_handle.fetch = mock.CoroutineMock(return_value=MockMasterSecret())
        secret = await self.holder.get_master_secret()
        assert secret is not None

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_get_master_secret_errors(self, mock_session_handle):
        # Not found
        mock_session_handle.fetch = mock.CoroutineMock(
            side_effect=[AskarError(code=AskarErrorCode.NOT_FOUND, message="test")]
        )
        with self.assertRaises(AnonCredsHolderError):
            await self.holder.get_master_secret()

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_get_master_secret_does_not_return_master_secret(
        self, mock_session_handle
    ):
        # Duplicate - Retry
        mock_session_handle.fetch = mock.CoroutineMock(return_value=None)
        mock_session_handle.insert = mock.CoroutineMock(
            side_effect=[AskarError(code=AskarErrorCode.DUPLICATE, message="test")]
        )
        with self.assertRaises(StopAsyncIteration):
            await self.holder.get_master_secret()
        # Other error
        mock_session_handle.fetch = mock.CoroutineMock(return_value=None)
        mock_session_handle.insert = mock.CoroutineMock(
            side_effect=[AskarError(code=AskarErrorCode.UNEXPECTED, message="test")]
        )
        with self.assertRaises(AnonCredsHolderError):
            await self.holder.get_master_secret()

    @mock.patch.object(AnonCredsHolder, "get_master_secret", return_value="master-secret")
    @mock.patch.object(
        CredentialRequest,
        "create",
        return_value=(mock.Mock(), mock.Mock()),
    )
    async def test_create_credential_request(
        self, mock_credential_request, mock_master_secret
    ):
        cred_def = CredDef(
            issuer_id="did:indy:sovrin:SGrjRL82Y9ZZbzhUDXokvQ",
            schema_id="did:indy:sovrin:SGrjRL82Y9ZZbzhUDXokvQ/anoncreds/v0/SCHEMA/MemberPass/1.0",
            tag="tag",
            type="CL",
            value=CredDefValue(primary=CredDefValuePrimary("n", "s", {}, "rctxt", "z")),
        )
        # error - to_native_fails
        with self.assertRaises(AnonCredsHolderError):
            await self.holder.create_credential_request(
                {"offer": "offer"},
                cred_def,
                "holder-did",
            )

        # Need to mock or else it will get ssl error
        cred_def.to_native = mock.MagicMock(return_value="native-cred-request")

        (
            cred_req_json,
            cred_req_metadata_json,
        ) = await self.holder.create_credential_request(
            {"offer": "offer"},
            cred_def,
            "holder-did",
        )

        assert cred_req_json is not None
        assert cred_req_metadata_json is not None
        assert mock_master_secret.called
        assert mock_credential_request.called

    async def test_create_credential_request_with_non_anoncreds_profile_throws_x(self):
        self.profile = await create_test_profile(
            settings={"wallet-type": "askar"},
        )
        self.holder = test_module.AnonCredsHolder(self.profile)
        with self.assertRaises(ValueError):
            await self.holder.create_credential_request(
                {"offer": "offer"},
                CredDef(
                    issuer_id="did:indy:sovrin:SGrjRL82Y9ZZbzhUDXokvQ",
                    schema_id="did:indy:sovrin:SGrjRL82Y9ZZbzhUDXokvQ/anoncreds/v0/SCHEMA/MemberPass/1.0",
                    tag="tag",
                    type="CL",
                    value=CredDefValue(
                        primary=CredDefValuePrimary("n", "s", {}, "rctxt", "z")
                    ),
                ),
                "holder-did",
            )

    @mock.patch.object(AnonCredsHolder, "get_master_secret", return_value="master-secret")
    async def test_store_credential_fails_to_load_raises_x(self, mock_master_secret):
        with self.assertRaises(AnonCredsHolderError):
            await self.holder.store_credential(
                {"cred-def": "cred-def"},
                {
                    "values": [
                        "name",
                        "date",
                        "degree",
                        "birthdate_dateint",
                        "timestamp",
                    ]
                },
                {"cred-req-meta": "cred-req-meta"},
            )

    @mock.patch.object(AnonCredsHolder, "get_master_secret", return_value="master-secret")
    @mock.patch.object(
        Credential,
        "load",
        side_effect=[
            MockCredential(),
            MockCredential(),
        ],
    )
    async def test_store_credential(self, mock_load, mock_master_secret):
        self.profile.transaction = mock.Mock(
            return_value=mock.MagicMock(
                insert=mock.CoroutineMock(return_value=None),
                commit=mock.CoroutineMock(return_value=None),
            )
        )
        self.profile.context.injector.bind_instance(
            AnonCredsRegistry, mock.MagicMock(AnonCredsRegistry, autospec=True)
        )

        # Valid
        result = await self.holder.store_credential(
            MOCK_CRED_DEF,
            MOCK_CRED,
            {"cred-req-meta": "cred-req-meta"},
            credential_attr_mime_types={"first_name": "application/json"},
        )

        assert result is not None
        assert mock_master_secret.called
        assert mock_load.called
        assert self.profile.transaction.called

        # Testing normalizing attr names
        await self.holder.store_credential(
            MOCK_CRED_DEF,
            {
                "values": {
                    "First Name": {"raw": "Alice", "encoded": "113...335"},
                },
            },
            {"cred-req-meta": "cred-req-meta"},
        )

    @mock.patch.object(AnonCredsHolder, "get_master_secret", return_value="master-secret")
    @mock.patch.object(
        W3cCredential,
        "load",
        side_effect=[
            MockW3Credential(),
        ],
    )
    @mock.patch.object(
        Credential,
        "from_w3c",
        side_effect=[
            MockCredential(),
        ],
    )
    async def test_store_credential_w3c(
        self, mock_load, mock_w3cload, mock_master_secret
    ):
        self.profile.context.injector.bind_instance(
            DocumentLoader, mock.MagicMock(DocumentLoader, autospec=True)
        )
        self.profile.transaction = mock.Mock(
            return_value=mock.MagicMock(
                insert=mock.CoroutineMock(return_value=None),
                commit=mock.CoroutineMock(return_value=None),
            )
        )
        self.profile.context.injector.bind_instance(
            AnonCredsRegistry, mock.MagicMock(AnonCredsRegistry, autospec=True)
        )
        with mock.patch.object(jsonld, "expand", return_value=MagicMock()):
            with mock.patch.object(JsonLdProcessor, "get_values", return_value=["type1"]):
                result = await self.holder.store_credential_w3c(
                    MOCK_CRED_DEF,
                    MOCK_W3C_CRED,
                    {"cred-req-meta": "cred-req-meta"},
                    credential_attr_mime_types={"first_name": "application/json"},
                )

                assert result is not None
                assert mock_master_secret.called
                assert mock_load.called
                assert mock_w3cload.called
                assert self.profile.transaction.called

    @mock.patch.object(AnonCredsHolder, "get_master_secret", return_value="master-secret")
    @mock.patch.object(Credential, "load", return_value=MockCredential())
    async def test_store_credential_failed_trx(self, *_):
        self.profile.transaction = mock.MagicMock(
            side_effect=[AskarError(AskarErrorCode.UNEXPECTED, "test")]
        )
        self.profile.context.injector.bind_instance(
            AnonCredsRegistry, mock.MagicMock(AnonCredsRegistry, autospec=True)
        )

        with self.assertRaises(AnonCredsHolderError):
            await self.holder.store_credential(
                MOCK_CRED_DEF,
                MOCK_CRED,
                {"cred-req-meta": "cred-req-meta"},
                credential_attr_mime_types={"first_name": "application/json"},
            )

    @mock.patch.object(Credential, "load", return_value=MockCredential())
    async def test_get_credentials(self, _):
        async with self.profile.session() as session:
            await session.handle.insert(CATEGORY_CREDENTIAL, json.dumps(MOCK_CRED))
        result = await self.holder.get_credentials(offset=0, limit=10, wql={})
        assert isinstance(result, list)
        assert len(result) == 1

    async def test_get_credentials_errors(self):
        self.profile.store.scan = mock.Mock(
            side_effect=[
                AskarError(AskarErrorCode.UNEXPECTED, "test"),
                AnoncredsError(AnoncredsErrorCode.UNEXPECTED, "test"),
            ]
        )

        with self.assertRaises(AnonCredsHolderError):
            await self.holder.get_credentials(offset=0, limit=10, wql={})
        with self.assertRaises(AnonCredsHolderError):
            await self.holder.get_credentials(offset=0, limit=10, wql={})

    async def test_get_credentials_for_presentation_request_by_referent(self):
        self.profile.store.scan = mock.Mock(
            return_value=mock.CoroutineMock(
                side_effect=[
                    MockCredScan(),
                ]
            )
        )

        # add predicates
        mock_pres_req = deepcopy(MOCK_PRES_REQ)
        mock_pres_req["requested_predicates"]["0_concentration_GE_uuid"] = {
            "name": "concentration",
            "p_type": "<=",
            "p_value": 9,
            "restrictions": [{"schema_name": "MYCO Biomarker"}],
        }
        await self.holder.get_credentials_for_presentation_request_by_referent(
            mock_pres_req, None, offset=0, limit=10
        )

        # non-existent referent
        with self.assertRaises(AnonCredsHolderError):
            await self.holder.get_credentials_for_presentation_request_by_referent(
                mock_pres_req, "not-found-ref", offset=0, limit=10
            )

    @mock.patch.object(Credential, "load", return_value=MockCredential())
    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_get_credential(self, mock_handle, _):
        mock_handle.fetch = mock.CoroutineMock(side_effect=[MockCredEntry(), None])
        result = await self.holder.get_credential("cred-id")
        assert isinstance(result, str)

        with self.assertRaises(WalletNotFoundError):
            await self.holder.get_credential("cred-id")

    @mock.patch.object(Credential, "load", return_value=MockCredential())
    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_credential_revoked(self, mock_handle, _):
        mock_registry = mock.MagicMock(AnonCredsRegistry, autospec=True)
        mock_registry.get_revocation_list = mock.CoroutineMock(
            return_value=GetRevListResult(
                revocation_list=RevList(
                    issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                    current_accumulator="21 124C594B6B20E41B681E92B2C43FD165EA9E68BC3C9D63A82C8893124983CAE94 21 124C5341937827427B0A3A32113BD5E64FB7AB39BD3E5ABDD7970874501CA4897 6 5438CB6F442E2F807812FD9DC0C39AFF4A86B1E6766DBB5359E86A4D70401B0F 4 39D1CA5C4716FFC4FE0853C4FF7F081DFD8DF8D2C2CA79705211680AC77BF3A1 6 70504A5493F89C97C225B68310811A41AD9CD889301F238E93C95AD085E84191 4 39582252194D756D5D86D0EED02BF1B95CE12AED2FA5CD3C53260747D891993C",
                    revocation_list=[0, 1, 1, 0],
                    timestamp=1669640864487,
                    rev_reg_def_id="rev-reg-id",
                ),
                resolution_metadata={},
                revocation_registry_metadata={},
            )
        )

        self.profile.context.injector.bind_instance(
            AnonCredsRegistry,
            mock_registry,
        )

        mock_handle.fetch = mock.CoroutineMock(return_value=MockCredEntry())
        assert (
            await self.holder.credential_revoked(
                credential_id="cred-id",
                timestamp_to=None,
                timestamp_from=None,
            )
            is False
        )

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_delete_credential(self, mock_handle):
        mock_handle.remove = mock.CoroutineMock(
            side_effect=[
                None,
                None,
                AskarError(AskarErrorCode.NOT_FOUND, "test"),
                AskarError(AskarErrorCode.UNEXPECTED, "test"),
            ]
        )
        await self.holder.delete_credential("cred-id")

        # not found
        with self.assertRaises(AnonCredsHolderError):
            await self.holder.delete_credential("cred-id")
        # other asker error, raise error
        with self.assertRaises(AnonCredsHolderError):
            await self.holder.delete_credential("cred-id")

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_get_mime_type(self, mock_handle):
        mock_handle.fetch = mock.CoroutineMock(
            side_effect=[
                MockMimeTypeRecord(),
                AskarError(AskarErrorCode.UNEXPECTED, "test"),
            ]
        )
        result = await self.holder.get_mime_type("cred-id", "mime-type")
        assert result == "mime-type"

        # asker error
        with self.assertRaises(AnonCredsHolderError):
            await self.holder.get_mime_type("cred-id", "mime-type")

    @mock.patch.object(Credential, "load", return_value=MockCredential())
    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    @mock.patch.object(AnonCredsHolder, "get_master_secret", return_value="master-secret")
    @mock.patch.object(
        anoncreds.Presentation, "create", return_value=Presentation.load(MOCK_PRES)
    )
    async def test_create_presentation(
        self, mock_pres_create, mock_master_secret, mock_handle, _
    ):
        mock_handle.fetch = mock.CoroutineMock(return_value=MockCredEntry())
        result = await self.holder.create_presentation(
            presentation_request=MOCK_PRES_REQ,
            requested_credentials={
                "self_attested_attributes": {},
                "requested_attributes": {},
                "requested_predicates": {},
            },
            schemas={},
            credential_definitions={},
            rev_states={},
        )

        json.loads(result)
        assert mock_pres_create.called
        assert mock_master_secret.called
        mock_handle.fetch.assert_called

        # requested_attributes and predicates
        await self.holder.create_presentation(
            presentation_request=MOCK_PRES_REQ,
            requested_credentials={
                "self_attested_attributes": {},
                "requested_attributes": {
                    "biomarker_attrs_0": {
                        "cred_id": "cred-id-requested",
                        "revealed": True,
                    },
                },
                "requested_predicates": {
                    "0_concentration_GE_uuid": {
                        "cred_id": "cred-id-predicate",
                    }
                },
            },
            schemas={},
            credential_definitions={},
            rev_states={},
        )

    @mock.patch.object(Credential, "load", return_value=MockCredential())
    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    @mock.patch.object(AnonCredsHolder, "get_master_secret", return_value="master-secret")
    @mock.patch.object(
        anoncreds.Presentation, "create", return_value=Presentation.load(MOCK_PRES)
    )
    @mock.patch.object(PresentCredentials, "add_attributes")
    async def test_create_presentation_with_revocation(
        self, mock_add_attributes, mock_pres_create, mock_master_secret, mock_handle, _
    ):
        mock_handle.fetch = mock.CoroutineMock(return_value=MockCredEntry(rev_reg=True))

        # not in rev_states
        with self.assertRaises(AnonCredsHolderError):
            await self.holder.create_presentation(
                presentation_request=MOCK_PRES_REQ,
                requested_credentials={
                    "self_attested_attributes": {},
                    "requested_attributes": {
                        "biomarker_attrs_0": {
                            "cred_id": "cred-id-requested",
                            "revealed": True,
                            "timestamp": 1234567890,
                        },
                    },
                    "requested_predicates": {
                        "0_concentration_GE_uuid": {
                            "cred_id": "cred-id-predicate",
                        }
                    },
                },
                schemas={},
                credential_definitions={},
                rev_states={},
            )
        # wrong timestamp
        with self.assertRaises(AnonCredsHolderError):
            await self.holder.create_presentation(
                presentation_request=MOCK_PRES_REQ,
                requested_credentials={
                    "self_attested_attributes": {},
                    "requested_attributes": {
                        "biomarker_attrs_0": {
                            "cred_id": "cred-id-requested",
                            "revealed": True,
                            "timestamp": 1234567890,
                        },
                    },
                    "requested_predicates": {
                        "0_concentration_GE_uuid": {
                            "cred_id": "cred-id-predicate",
                        }
                    },
                },
                schemas={},
                credential_definitions={},
                rev_states={
                    "rev-reg-id": {
                        "9999999999": b"100",
                    }
                },
            )

        await self.holder.create_presentation(
            presentation_request=MOCK_PRES_REQ,
            requested_credentials={
                "self_attested_attributes": {},
                "requested_attributes": {
                    "biomarker_attrs_0": {
                        "cred_id": "cred-id-requested",
                        "revealed": True,
                        "timestamp": 1234567890,
                    },
                },
                "requested_predicates": {
                    "0_concentration_GE_uuid": {
                        "cred_id": "cred-id-predicate",
                    }
                },
            },
            schemas={},
            credential_definitions={},
            rev_states={
                "rev-reg-id": {
                    1234567890: b'{"witness":{"omega": "21 124...AC8"}}',
                }
            },
        )

        assert mock_add_attributes.called
        assert mock_pres_create.called
        assert mock_master_secret.called
        assert mock_handle.fetch.called

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    @mock.patch.object(AnonCredsHolder, "get_master_secret", return_value="master-secret")
    @mock.patch.object(
        anoncreds.Presentation,
        "create",
        side_effect=AnoncredsError(AnoncredsErrorCode.UNEXPECTED, "test"),
    )
    async def test_create_presentation_create_error(
        self, mock_pres_create, mock_master_secret, mock_handle
    ):
        mock_handle.fetch = mock.CoroutineMock(return_value=MockCredEntry())
        # anoncreds error when creating presentation
        with self.assertRaises(AnonCredsHolderError):
            await self.holder.create_presentation(
                presentation_request=MOCK_PRES_REQ,
                requested_credentials={
                    "self_attested_attributes": {},
                    "requested_attributes": {},
                    "requested_predicates": {},
                },
                schemas={},
                credential_definitions={},
                rev_states={},
            )

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    @mock.patch.object(AnonCredsHolder, "get_master_secret", return_value="master-secret")
    @mock.patch.object(
        anoncreds.W3cPresentation,
        "create",
        return_value=W3cPresentation.load(MOCK_W3CPRES),
    )
    async def test_create_presentation_w3c(
        self, mock_pres_create, mock_master_secret, mock_handle
    ):
        mock_handle.fetch = mock.CoroutineMock(return_value=MockCredEntry())
        await self.holder.create_presentation_w3c(
            presentation_request=MOCK_W3C_PRES_REQ,
            requested_credentials_w3c=[W3cCredential.load(MOCK_W3C_CRED)],
            schemas=SCHEMAS,
            credential_definitions=CRED_DEFS,
            credentials_w3c_metadata=CREDS_W3C_METADATA,
        )

        assert mock_pres_create.called
        assert mock_master_secret.called
        mock_handle.fetch.assert_called

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    @mock.patch.object(AnonCredsHolder, "get_master_secret", return_value="master-secret")
    @mock.patch.object(
        anoncreds.W3cPresentation,
        "create",
        side_effect=AnoncredsError(AnoncredsErrorCode.UNEXPECTED, "test"),
    )
    async def test_create_presentation_w3c_create_error(
        self, mock_pres_create, mock_master_secret, mock_handle
    ):
        mock_handle.fetch = mock.CoroutineMock(return_value=MockCredEntry())
        # anoncreds error when creating presentation
        with self.assertRaises(AnonCredsHolderError):
            await self.holder.create_presentation_w3c(
                presentation_request=MOCK_W3C_PRES_REQ,
                requested_credentials_w3c=[W3cCredential.load(MOCK_W3C_CRED)],
                schemas=SCHEMAS,
                credential_definitions=CRED_DEFS,
                credentials_w3c_metadata=CREDS_W3C_METADATA,
            )

    @mock.patch.object(
        CredentialRevocationState,
        "create",
    )
    async def test_create_revocation_state(self, mock_create):
        schema = Schema.create(
            name="MemberPass",
            attr_names=["member", "score"],
            issuer_id="did:indy:sovrin:SGrjRL82Y9ZZbzhUDXokvQ",
            version="1.0",
        )

        (cred_def, _, _) = CredentialDefinition.create(
            schema_id="did:indy:sovrin:SGrjRL82Y9ZZbzhUDXokvQ/anoncreds/v0/SCHEMA/MemberPass/1.0",
            schema=schema,
            issuer_id="did:indy:sovrin:SGrjRL82Y9ZZbzhUDXokvQ",
            tag="tag",
            support_revocation=True,
            signature_type="CL",
        )
        (rev, _) = RevocationRegistryDefinition.create(
            cred_def_id="SGrjRL82Y9ZZbzhUDXokvQ:3:CL:531757:MemberPass",
            cred_def=cred_def,
            issuer_id="did:indy:sovrin:SGrjRL82Y9ZZbzhUDXokvQ",
            registry_type="CL_ACCUM",
            max_cred_num=100,
            tag="tag",
        )
        mock_create.return_value = rev
        result = await self.holder.create_revocation_state(
            cred_rev_id="1",
            rev_reg_def={"def": 1},
            rev_list={"accum": "1"},
            tails_file_path="/tmp/some.tails",
        )

        result = json.loads(result)
        assert mock_create.called

        # error
        mock_create.side_effect = AnoncredsError(AnoncredsErrorCode.UNEXPECTED, "test")
        with self.assertRaises(AnonCredsHolderError):
            await self.holder.create_revocation_state(
                cred_rev_id="1",
                rev_reg_def={"def": 1},
                rev_list={"accum": "1"},
                tails_file_path="/tmp/some.tails",
            )



"""Peer DID Resolver.

Resolution is performed by converting did:peer:2 to did:peer:3 according to
https://identity.foundation/peer-did-method-spec/#generation-method:~:text=Method%203%3A%20DID%20Shortening%20with%20SHA%2D256%20Hash
"""

import logging
import re
from typing import Optional, Pattern, Sequence, Text

from did_peer_2 import PATTERN as PEER2_PATTERN
from did_peer_2 import PEER3_PATTERN, peer2to3, resolve_peer3

from ...config.injection_context import InjectionContext
from ...core.event_bus import Event, EventBus
from ...core.profile import Profile
from ...storage.base import BaseStorage
from ...storage.error import StorageNotFoundError
from ...storage.record import StorageRecord
from ..base import BaseDIDResolver, DIDNotFound, ResolverType

LOGGER = logging.getLogger(__name__)


class PeerDID3Resolver(BaseDIDResolver):
    """Peer DID Resolver."""

    RECORD_TYPE_3_TO_2 = "peer3_to_peer2"

    def __init__(self):
        """Initialize Key Resolver."""
        super().__init__(ResolverType.NATIVE)

    async def setup(self, context: InjectionContext):
        """Perform required setup for Key DID resolution."""
        event_bus = context.inject(EventBus)
        event_bus.subscribe(
            re.compile("acapy::record::connections::deleted"),
            self.remove_record_for_deleted_conn,
        )

    @property
    def supported_did_regex(self) -> Pattern:
        """Return supported_did_regex of Key DID Resolver."""
        return PEER3_PATTERN

    async def _resolve(
        self,
        profile: Profile,
        did: str,
        service_accept: Optional[Sequence[Text]] = None,
    ) -> dict:
        """Resolve a did:peer:3 DID."""
        async with profile.session() as session:
            storage = session.inject(BaseStorage)
            try:
                record = await storage.get_record(self.RECORD_TYPE_3_TO_2, did)
            except StorageNotFoundError:
                raise DIDNotFound(
                    f"did:peer:3 does not correspond to a known did:peer:2 {did}"
                )

        doc = resolve_peer3(record.value)
        return doc

    async def create_and_store(self, profile: Profile, peer2: str):
        """Inject did:peer:2 create did:peer:3 and store document."""
        if not PEER2_PATTERN.match(peer2):
            raise ValueError("did:peer:2 expected")

        peer3 = peer2to3(peer2)
        async with profile.session() as session:
            storage = session.inject(BaseStorage)
            try:
                record = await storage.get_record(self.RECORD_TYPE_3_TO_2, peer3)
            except StorageNotFoundError:
                record = StorageRecord(self.RECORD_TYPE_3_TO_2, peer2, {}, peer3)
                await storage.add_record(record)
            else:
                pass

        doc = resolve_peer3(peer2)
        return doc

    async def remove_record_for_deleted_conn(self, profile: Profile, event: Event):
        """Remove record for deleted connection, if found."""
        their_did = event.payload.get("their_did")
        my_did = event.payload.get("my_did")
        if not their_did and not my_did:
            return
        dids = [
            *(did for did in (their_did, my_did) if did and PEER3_PATTERN.match(did)),
            *(
                peer2to3(did)
                for did in (their_did, my_did)
                if did and PEER2_PATTERN.match(did)
            ),
        ]
        if dids:
            LOGGER.debug("Removing peer 2 to 3 mapping for deleted connection: %s", dids)
        async with profile.session() as session:
            storage = session.inject(BaseStorage)
            for did in dids:
                try:
                    record = StorageRecord(self.RECORD_TYPE_3_TO_2, None, None, did)
                    await storage.delete_record(record)
                except StorageNotFoundError:
                    LOGGER.debug("No peer 2 to 3 mapping found for: %s", did)



"""Record used to represent the form associated with an action menu option."""

from typing import Optional, Sequence

from marshmallow import EXCLUDE, fields

from .....messaging.models.base import BaseModel, BaseModelSchema
from .menu_form_param import MenuFormParam, MenuFormParamSchema


class MenuForm(BaseModel):
    """Instance of a form associated with an action menu item."""

    class Meta:
        """Menu form metadata."""

        schema_class = "MenuFormSchema"

    def __init__(
        self,
        *,
        title: Optional[str] = None,
        description: Optional[str] = None,
        params: Sequence[MenuFormParam] = None,
        submit_label: Optional[str] = None,
    ):
        """Initialize a MenuForm instance.

        Args:
            title: The menu form title
            description: Additional descriptive text for the menu form
            params: A list of form parameters
            submit_label: An alternative label for the form submit button
        """
        self.title = title
        self.description = description
        self.params = list(params) if params else []
        self.submit_label = submit_label


class MenuFormSchema(BaseModelSchema):
    """MenuForm schema."""

    class Meta:
        """MenuFormSchema metadata."""

        model_class = MenuForm
        unknown = EXCLUDE

    title = fields.Str(
        required=False,
        metadata={"description": "Menu form title", "example": "Preferences"},
    )
    description = fields.Str(
        required=False,
        metadata={
            "description": "Additional descriptive text for menu form",
            "example": "Window preference settings",
        },
    )
    params = fields.List(
        fields.Nested(MenuFormParamSchema()),
        required=False,
        metadata={"description": "List of form parameters"},
    )
    submit_label = fields.Str(
        required=False,
        data_key="submit-label",
        metadata={
            "description": "Alternative label for form submit button",
            "example": "Send",
        },
    )



"""Inbound connection handling classes."""

import asyncio
import logging
from typing import Callable, Optional, Sequence, Union

from ...admin.server import AdminResponder
from ...core.profile import Profile
from ...messaging.responder import BaseResponder
from ...multitenant.base import BaseMultitenantManager
from ..error import WireFormatError
from ..outbound.message import OutboundMessage
from ..wire_format import BaseWireFormat
from .message import InboundMessage
from .receipt import MessageReceipt

LOGGER = logging.getLogger(__name__)


class AcceptResult:
    """Represent the result of accept_response."""

    def __init__(self, accepted: bool, retry: bool = False):
        """Initialize the `AcceptResult` instance."""
        self.accepted = accepted
        self.retry = retry

    def __bool__(self) -> bool:
        """Check if the result is true."""
        return self.accepted


class InboundSession:
    """Track an open transport connection for direct routing of outbound messages."""

    def __init__(
        self,
        *,
        profile: Profile,
        inbound_handler: Callable,
        session_id: str,
        wire_format: BaseWireFormat,
        accept_undelivered: bool = False,
        can_respond: bool = False,
        client_info: Optional[dict] = None,
        close_handler: Optional[Callable] = None,
        reply_mode: Optional[str] = None,
        reply_thread_ids: Sequence[str] = None,
        reply_verkeys: Sequence[str] = None,
        transport_type: Optional[str] = None,
    ):
        """Initialize the inbound session."""
        self.profile = profile
        self.inbound_handler = inbound_handler
        self.session_id = session_id
        self.wire_format = wire_format

        self.accept_undelivered = accept_undelivered
        self.client_info = client_info
        self.close_handler = close_handler
        self.response_buffer: Optional[OutboundMessage] = None
        self.response_event = asyncio.Event()
        self.transport_type = transport_type

        self._can_respond = can_respond
        self._closed = False
        self._reply_mode = None
        self._reply_verkeys = None
        self._reply_thread_ids = None

        # If multitenancy is enabled we need to relay the message by changing
        # the context/profile to the wallet associated with the message.
        # Only needs to happen for the first received message
        self._check_relay_context = profile.context.settings.get(
            "multitenant.enabled", False
        )

        # call setters
        self.reply_thread_ids = reply_thread_ids
        self.reply_verkeys = reply_verkeys
        self.reply_mode = reply_mode

    @property
    def can_respond(self) -> bool:
        """Accessor for the session can-respond state."""
        return self._can_respond and not self._closed

    @can_respond.setter
    def can_respond(self, can_respond: bool):
        """Setter for the session can-respond state."""
        self._can_respond = can_respond

    @property
    def closed(self) -> bool:
        """Accessor for the session closed state."""
        return self._closed

    def close(self):
        """Setter for the session closed state."""
        self._closed = True
        self.response_event.set()  # end wait_response if blocked
        if self.close_handler:
            self.close_handler(self)

    @property
    def reply_mode(self) -> str:
        """Accessor for the session reply mode."""
        return self._reply_mode

    @reply_mode.setter
    def reply_mode(self, mode: str):
        """Setter for the session reply mode."""
        if mode not in (
            MessageReceipt.REPLY_MODE_ALL,
            MessageReceipt.REPLY_MODE_THREAD,
        ):
            mode = None
        self._reply_mode = mode
        if not mode:
            # reset the tracked thread IDs when the mode is changed to none
            self.reply_thread_ids = set()

    @property
    def reply_verkeys(self):
        """Accessor for the reply verkeys."""
        return self._reply_verkeys.copy()

    @reply_verkeys.setter
    def reply_verkeys(self, verkeys: Sequence[str]):
        """Setter for the reply verkeys."""
        self._reply_verkeys = set(verkeys) if verkeys else set()

    @property
    def reply_thread_ids(self):
        """Accessor for the reply thread IDs."""
        return self._reply_thread_ids.copy()

    @reply_thread_ids.setter
    def reply_thread_ids(self, thread_ids: Sequence[str]):
        """Setter for the reply thread IDs."""
        self._reply_thread_ids = set(thread_ids) if thread_ids else set()

    def add_reply_thread_ids(self, *thids):
        """Add a thread ID to the set of potential reply targets."""
        for thid in filter(None, thids):
            self._reply_thread_ids.add(thid)

    def add_reply_verkeys(self, *verkeys):
        """Add a verkey to the set of potential reply targets."""
        for verkey in filter(None, verkeys):
            self._reply_verkeys.add(verkey)

    @property
    def response_buffered(self) -> bool:
        """Check if a response is currently buffered."""
        return bool(self.response_buffer)

    async def handle_relay_context(self, payload_enc: Union[str, bytes]):
        """Update the session profile based on the recipients of an incoming message."""
        multitenant_mgr = self.profile.context.inject(BaseMultitenantManager)

        try:
            [wallet] = await multitenant_mgr.get_wallets_by_message(
                payload_enc, self.wire_format
            )

            if wallet.is_managed:
                profile = await multitenant_mgr.get_wallet_profile(
                    self.profile.context, wallet
                )

                base_responder: AdminResponder = profile.inject(BaseResponder)

                # Create new responder based on base responder
                responder = AdminResponder(
                    profile,
                    base_responder.send_fn,
                )
                profile.context.injector.bind_instance(BaseResponder, responder)

                # overwrite session profile with wallet profile
                self.profile = profile

        except ValueError:
            pass  # No wallet found. Use the base session profile

    def process_inbound(self, message: InboundMessage):
        """Process an incoming message and update the session metadata as necessary.

        Args:
            message: The inbound message instance
        """
        receipt = message.receipt
        mode = self.reply_mode = (
            receipt.direct_response_requested and receipt.direct_response_mode
        )
        self.add_reply_verkeys(receipt.sender_verkey)
        if mode == MessageReceipt.REPLY_MODE_THREAD:
            self.add_reply_thread_ids(receipt.thread_id)

    async def parse_inbound(self, payload_enc: Union[str, bytes]) -> InboundMessage:
        """Convert a message payload and to an inbound message."""
        session = await self.profile.session()
        payload, receipt = await self.wire_format.parse_message(session, payload_enc)
        return InboundMessage(
            payload,
            receipt,
            session_id=self.session_id,
            transport_type=self.transport_type,
        )

    async def receive(self, payload_enc: Union[str, bytes]) -> InboundMessage:
        """Receive a new message payload and dispatch the message."""
        if self._check_relay_context:
            await self.handle_relay_context(payload_enc)
            self._check_relay_context = False

        message = await self.parse_inbound(payload_enc)
        self.receive_inbound(message)
        return message

    def receive_inbound(self, message: InboundMessage):
        """Deliver the inbound message to the conductor."""
        self.process_inbound(message)
        self.inbound_handler(self.profile, message, can_respond=self.can_respond)

    def select_outbound(self, message: OutboundMessage) -> bool:
        """Determine if an outbound message should be sent to this session.

        Args:
            message: The outbound message to be checked

        """
        if not self.can_respond:
            return False

        mode = self.reply_mode
        reply_verkey = message.reply_to_verkey
        reply_thread_id = message.reply_thread_id

        if reply_verkey and reply_verkey in self.reply_verkeys:
            if mode == MessageReceipt.REPLY_MODE_ALL:
                return True
            elif (
                mode == MessageReceipt.REPLY_MODE_THREAD
                and reply_thread_id
                and reply_thread_id in self._reply_thread_ids
            ):
                return True

        return False

    async def encode_outbound(self, outbound: OutboundMessage) -> OutboundMessage:
        """Apply wire formatting to an outbound message."""
        if not outbound.payload:
            raise WireFormatError("Message has no payload to encode")
        if not outbound.reply_to_verkey:
            raise WireFormatError("No reply verkey available for encoding message")

        session = await self.profile.session()
        return await self.wire_format.encode_message(
            session,
            outbound.payload,
            [outbound.reply_to_verkey],
            None,
            outbound.reply_from_verkey,
        )

    def accept_response(self, message: OutboundMessage) -> AcceptResult:
        """Try to queue an outbound message if it applies to this session.

        Returns: a tuple of (message buffered, retry later)
        """
        if not self.select_outbound(message):
            return AcceptResult(False, False)
        if self.response_buffer:
            return AcceptResult(False, True)
        self.set_response(message)
        return AcceptResult(True)

    def set_response(self, message: OutboundMessage):
        """Set the contents of the response message buffer."""
        self.response_buffer = message
        self.response_event.set()

    def clear_response(self):
        """Handle when the buffered response message has been delivered."""
        self.response_buffer = None
        self.response_event.set()

    async def wait_response(self) -> Union[str, bytes]:
        """Wait for a response to be buffered and pack it."""
        while True:
            if self._closed:
                return
            if self.response_buffer:
                response = self.response_buffer.enc_payload
                if not response:
                    try:
                        response = await self.encode_outbound(self.response_buffer)
                    except WireFormatError as e:
                        LOGGER.warning("Error encoding direct response: %s", str(e))
                        self.clear_response()
                if response:
                    return response
            self.response_event.clear()
            await self.response_event.wait()

    async def __aenter__(self):
        """Async context manager entry."""
        return self

    async def __aexit__(self, exc_type, exc_value, exc_tb):
        """Async context manager entry."""
        self.close()



"""Websockets outbound transport."""

import logging
from typing import Optional, Union

from aiohttp import ClientSession, DummyCookieJar

from ...core.profile import Profile
from .base import BaseOutboundTransport


class WsTransport(BaseOutboundTransport):
    """Websockets outbound transport class."""

    schemes = ("ws", "wss")
    is_external = False

    def __init__(self, **kwargs) -> None:
        """Initialize an `WsTransport` instance."""
        super().__init__(**kwargs)
        self.logger = logging.getLogger(__name__)

    async def start(self):
        """Start the outbound transport."""
        self.client_session = ClientSession(cookie_jar=DummyCookieJar(), trust_env=True)
        return self

    async def stop(self):
        """Stop the outbound transport."""
        await self.client_session.close()
        self.client_session = None

    async def handle_message(
        self,
        profile: Profile,
        payload: Union[str, bytes],
        endpoint: str,
        metadata: Optional[dict] = None,
        api_key: Optional[str] = None,
    ):
        """Handle message from queue.

        Args:
            profile: the profile that produced the message
            payload: message payload in string or byte format
            endpoint: URI endpoint for delivery
            metadata: Additional metadata associated with the payload
            api_key: API key for the endpoint
        """
        # aiohttp should automatically handle websocket sessions
        async with self.client_session.ws_connect(endpoint, headers=metadata) as ws:
            self.logger.debug("Sending outbound websocket message %s", payload)
            if isinstance(payload, bytes):
                await ws.send_bytes(payload)
            else:
                await ws.send_str(payload)



"""Wallet utility functions."""

import base64
import re
from typing import Optional

import base58
import nacl.bindings
import nacl.utils

from ..core.profile import Profile


def random_seed() -> bytes:
    """Generate a random seed value.

    Returns:
        A new random seed

    """
    return nacl.utils.random(nacl.bindings.crypto_box_SEEDBYTES)


def pad(val: str) -> str:
    """Pad base64 values if need be: JWT calls to omit trailing padding."""
    padlen = 4 - len(val) % 4
    return val if padlen > 2 else (val + "=" * padlen)


def unpad(val: str) -> str:
    """Remove padding from base64 values if need be."""
    return val.rstrip("=")


def b64_to_bytes(val: str, urlsafe=False) -> bytes:
    """Convert a base 64 string to bytes."""
    if urlsafe:
        return base64.urlsafe_b64decode(pad(val))
    return base64.b64decode(pad(val))


def b64_to_str(val: str, urlsafe=False, encoding=None) -> str:
    """Convert a base 64 string to string on input encoding (default utf-8)."""
    return b64_to_bytes(val, urlsafe).decode(encoding or "utf-8")


def bytes_to_b64(val: bytes, urlsafe=False, pad=True, encoding: str = "ascii") -> str:
    """Convert a byte string to base 64."""
    b64 = (
        base64.urlsafe_b64encode(val).decode(encoding)
        if urlsafe
        else base64.b64encode(val).decode(encoding)
    )
    return b64 if pad else unpad(b64)


def str_to_b64(val: str, urlsafe=False, encoding=None, pad=True) -> str:
    """Convert a string to base64 string on input encoding (default utf-8)."""
    return bytes_to_b64(val.encode(encoding or "utf-8"), urlsafe, pad)


def set_urlsafe_b64(val: str, urlsafe: bool = True) -> str:
    """Set URL safety in base64 encoding."""
    if urlsafe:
        return val.replace("+", "-").replace("/", "_")
    return val.replace("-", "+").replace("_", "/")


def b58_to_bytes(val: str) -> bytes:
    """Convert a base 58 string to bytes."""
    return base58.b58decode(val)


def bytes_to_b58(val: bytes) -> str:
    """Convert a byte string to base 58."""
    return base58.b58encode(val).decode("ascii")


def full_verkey(did: str, abbr_verkey: str) -> str:
    """Given a DID and abbreviated verkey, return the full verkey."""
    return (
        bytes_to_b58(b58_to_bytes(did.split(":")[-1]) + b58_to_bytes(abbr_verkey[1:]))
        if abbr_verkey.startswith("~")
        else abbr_verkey
    )


def default_did_from_verkey(verkey: str) -> str:
    """Given a verkey, return the default indy did.

    By default the did is the first 16 bytes of the verkey.
    """
    did = bytes_to_b58(b58_to_bytes(verkey)[:16])
    return did


def abbr_verkey(full_verkey: str, did: Optional[str] = None) -> str:
    """Given a full verkey and DID, return the abbreviated verkey."""
    did_len = len(b58_to_bytes(did.split(":")[-1])) if did else 16
    return f"~{bytes_to_b58(b58_to_bytes(full_verkey)[did_len:])}"


DID_EVENT_PREFIX = "acapy::ENDORSE_DID::"
DID_ATTRIB_EVENT_PREFIX = "acapy::ENDORSE_DID_ATTRIB::"
EVENT_LISTENER_PATTERN = re.compile(f"^{DID_EVENT_PREFIX}(.*)?$")
ATTRIB_EVENT_LISTENER_PATTERN = re.compile(f"^{DID_ATTRIB_EVENT_PREFIX}(.*)?$")


async def notify_endorse_did_event(profile: Profile, did: str, meta_data: dict):
    """Send notification for a DID post-process event."""
    await profile.notify(
        DID_EVENT_PREFIX + did,
        meta_data,
    )


async def notify_endorse_did_attrib_event(profile: Profile, did: str, meta_data: dict):
    """Send notification for a DID ATTRIB post-process event."""
    await profile.notify(
        DID_ATTRIB_EVENT_PREFIX + did,
        meta_data,
    )



"""Represents a DID exchange request message under RFC 23."""

from typing import Optional

from marshmallow import EXCLUDE, fields

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from .....messaging.decorators.attach_decorator import (
    AttachDecorator,
    AttachDecoratorSchema,
)
from .....messaging.valid import GENERIC_DID_EXAMPLE, GENERIC_DID_VALIDATE
from ..message_types import DIDX_REQUEST, PROTOCOL_PACKAGE

HANDLER_CLASS = f"{PROTOCOL_PACKAGE}.handlers.request_handler.DIDXRequestHandler"


class DIDXRequest(AgentMessage):
    """Class representing a DID exchange request under RFC 23."""

    class Meta:
        """Metadata for DID exchange request under RFC 23."""

        handler_class = HANDLER_CLASS
        message_type = DIDX_REQUEST
        schema_class = "DIDXRequestSchema"

    def __init__(
        self,
        *,
        label: Optional[str] = None,
        did: Optional[str] = None,
        did_doc_attach: Optional[AttachDecorator] = None,
        goal_code: Optional[str] = None,
        goal: Optional[str] = None,
        **kwargs,
    ):
        """Initialize DID exchange request object under RFC 23.

        Args:
            label: Label for this request
            did: DID for this request
            did_doc_attach: signed DID doc attachment
            goal_code: (optional) is a self-attested code the receiver may want to
              display to the user or use in automatically deciding what to do with
              the request message. The goal code might be used particularly when the
              request is sent to a resolvable DID without reference to a specific
              invitation.
            goal: (optional) is a self-attested string that the receiver may want to
              display to the user about the context-specific goal of the request message.
            kwargs: Additional keyword arguments for the message
        """
        super().__init__(**kwargs)
        self.label = label
        self.did = did
        self.did_doc_attach = did_doc_attach
        self.goal_code = goal_code
        self.goal = goal


class DIDXRequestSchema(AgentMessageSchema):
    """Schema class for DID exchange request under RFC 23."""

    class Meta:
        """DID exchange request schema class metadata."""

        model_class = DIDXRequest
        unknown = EXCLUDE

    label = fields.Str(
        required=True,
        metadata={
            "description": "Label for DID exchange request",
            "example": "Request to connect with Bob",
        },
    )
    did = fields.Str(
        validate=GENERIC_DID_VALIDATE,
        metadata={"description": "DID of exchange", "example": GENERIC_DID_EXAMPLE},
    )
    did_doc_attach = fields.Nested(
        AttachDecoratorSchema,
        required=False,
        data_key="did_doc~attach",
        metadata={"description": "As signed attachment, DID Doc associated with DID"},
    )
    goal_code = fields.Str(
        required=False,
        metadata={
            "description": (
                "A self-attested code the receiver may want to display to the user or"
                " use in automatically deciding what to do with the out-of-band message"
            ),
            "example": "issue-vc",
        },
    )
    goal = fields.Str(
        required=False,
        metadata={
            "description": (
                "A self-attested string that the receiver may want to display to the"
                " user about the context-specific goal of the out-of-band message"
            ),
            "example": "To issue a Faber College Graduate credential",
        },
    )



DID_zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa = {
    "@context": "https://www.w3.org/ns/did/v1",
    "id": "did:key:zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa",
    "verificationMethod": [
        {
            "id": "did:key:zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa#zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa",
            "type": "Bls12381G2Key2020",
            "controller": "did:key:zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa",
            "publicKeyBase58": "nZZe9Nizhaz9JGpgjysaNkWGg5TNEhpib5j6WjTUHJ5K46dedUrZ57PUFZBq9Xckv8mFJjx6G6Vvj2rPspq22BagdADEEEy2F8AVLE1DhuwWC5vHFa4fUhUwxMkH7B6joqG",
        }
    ],
    "authentication": [
        "did:key:zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa#zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa"
    ],
    "assertionMethod": [
        "did:key:zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa#zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa"
    ],
    "capabilityDelegation": [
        "did:key:zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa#zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa"
    ],
    "capabilityInvocation": [
        "did:key:zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa#zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa"
    ],
    "keyAgreement": [],
}



"""Represents a Transaction Job to send message."""

from typing import Optional

from marshmallow import EXCLUDE, fields

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from ..message_types import PROTOCOL_PACKAGE, TRANSACTION_JOB_TO_SEND

HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers"
    ".transaction_job_to_send_handler.TransactionJobToSendHandler"
)


class TransactionJobToSend(AgentMessage):
    """Class representing a transaction job to send."""

    class Meta:
        """Metadata for a TransactionJobToSend."""

        handler_class = HANDLER_CLASS
        message_type = TRANSACTION_JOB_TO_SEND
        schema_class = "TransactionJobToSendSchema"

    def __init__(
        self,
        *,
        job: Optional[str] = None,
        **kwargs,
    ):
        """Initialize transaction job to send.

        Args:
            job: The job that needs to be send
            kwargs: Additional keyword arguments for the message

        """

        super().__init__(**kwargs)
        self.job = job


class TransactionJobToSendSchema(AgentMessageSchema):
    """Transaction Job to send schema class."""

    class Meta:
        """Metadata for a TransactionJobToSendSchema."""

        model_class = TransactionJobToSend
        unknown = EXCLUDE

    job = fields.Str(
        required=True,
        metadata={
            "description": "Transaction job that is sent to the other agent",
            "example": "TRANSACTION_AUTHOR",
        },
    )



"""Version definitions for this protocol."""

versions = [
    {
        "major_version": 1,
        "minimum_minor_version": 0,
        "current_minor_version": 0,
        "path": "v1_0",
    },
    {
        "major_version": 2,
        "minimum_minor_version": 0,
        "current_minor_version": 0,
        "path": "v2_0",
    },
]



"""Utility functions for server operations in an AcaPy agent."""


async def remove_unwanted_headers(
    request,
    response,
) -> None:
    """Remove unwanted headers from the response."""
    response.headers.pop("Server", None)



from unittest import IsolatedAsyncioTestCase

from .....admin.request_context import AdminRequestContext
from .....connections.models.conn_record import ConnRecord
from .....tests import mock
from .....utils.testing import create_test_profile
from .. import routes as test_module


class TestOutOfBandRoutes(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile(
            settings={
                "admin.admin_api_key": "secret-key",
            }
        )
        self.context = AdminRequestContext.test_context({}, profile=self.profile)
        self.request_dict = {
            "context": self.context,
            "outbound_message_router": mock.CoroutineMock(),
        }
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            headers={"x-api-key": "secret-key"},
        )

    async def test_invitation_create(self):
        self.request.query = {
            "multi_use": "true",
            "auto_accept": "true",
        }
        body = {
            "attachments": mock.MagicMock(),
            "handshake_protocols": [test_module.HSProto.RFC23.name],
            "use_public_did": True,
            "metadata": {"hello": "world"},
        }
        self.request.json = mock.CoroutineMock(return_value=body)

        with (
            mock.patch.object(
                test_module, "OutOfBandManager", autospec=True
            ) as mock_oob_mgr,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_oob_mgr.return_value.create_invitation = mock.CoroutineMock(
                return_value=mock.MagicMock(
                    serialize=mock.MagicMock(return_value={"abc": "123"})
                )
            )

            await test_module.invitation_create(self.request)
            mock_oob_mgr.return_value.create_invitation.assert_called_once_with(
                my_label=None,
                auto_accept=True,
                public=True,
                use_did_method=None,
                use_did=None,
                multi_use=True,
                create_unique_did=False,
                hs_protos=[test_module.HSProto.RFC23],
                attachments=body["attachments"],
                metadata=body["metadata"],
                alias=None,
                mediation_id=None,
                service_accept=None,
                protocol_version=None,
                goal_code=None,
                goal=None,
            )
            mock_json_response.assert_called_once_with({"abc": "123"})

    async def test_invitation_fetch(self):
        self.request.query = {"oob_id": "dummy"}

        with (
            mock.patch.object(
                test_module, "OutOfBandManager", autospec=True
            ) as mock_oob_mgr,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_oob_mgr.return_value.fetch_oob_invitation_record_by_id = (
                mock.CoroutineMock(
                    return_value=mock.MagicMock(
                        serialize=mock.MagicMock(return_value={"abc": "123"})
                    )
                )
            )
            await test_module.invitation_fetch(self.request)
            mock_json_response.assert_called_once_with({"abc": "123"})

    async def test_invitation_remove(self):
        self.request.match_info = {"invi_msg_id": "dummy"}

        with (
            mock.patch.object(
                test_module, "OutOfBandManager", autospec=True
            ) as mock_oob_mgr,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_oob_mgr.return_value.delete_conn_and_oob_record_invitation = (
                mock.CoroutineMock(return_value=None)
            )
            await test_module.invitation_remove(self.request)
            mock_json_response.assert_called_once_with({})

    async def test_invitation_create_with_accept(self):
        self.request.query = {
            "multi_use": "true",
            "auto_accept": "true",
        }
        body = {
            "attachments": mock.MagicMock(),
            "handshake_protocols": [test_module.HSProto.RFC23.name],
            "accept": ["didcomm/aip1", "didcomm/aip2;env=rfc19"],
            "use_public_did": True,
            "metadata": {"hello": "world"},
        }
        self.request.json = mock.CoroutineMock(return_value=body)

        with (
            mock.patch.object(
                test_module, "OutOfBandManager", autospec=True
            ) as mock_oob_mgr,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_oob_mgr.return_value.create_invitation = mock.CoroutineMock(
                return_value=mock.MagicMock(
                    serialize=mock.MagicMock(return_value={"abc": "123"})
                )
            )

            await test_module.invitation_create(self.request)
            mock_oob_mgr.return_value.create_invitation.assert_called_once_with(
                my_label=None,
                auto_accept=True,
                public=True,
                use_did_method=None,
                use_did=None,
                multi_use=True,
                create_unique_did=False,
                hs_protos=[test_module.HSProto.RFC23],
                attachments=body["attachments"],
                metadata=body["metadata"],
                alias=None,
                mediation_id=None,
                service_accept=["didcomm/aip1", "didcomm/aip2;env=rfc19"],
                protocol_version=None,
                goal_code=None,
                goal=None,
            )
            mock_json_response.assert_called_once_with({"abc": "123"})

    async def test_invitation_create_x(self):
        self.request.query = {"multi_use": "true"}
        self.request.json = mock.CoroutineMock(
            return_value={
                "attachments": mock.MagicMock(),
                "handshake_protocols": [23],
                "use_public_did": True,
            }
        )

        with (
            mock.patch.object(
                test_module, "OutOfBandManager", autospec=True
            ) as mock_oob_mgr,
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_oob_mgr.return_value.create_invitation = mock.CoroutineMock(
                side_effect=test_module.OutOfBandManagerError()
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.invitation_create(self.request)
            mock_json_response.assert_not_called()

    async def test_invitation_receive(self):
        self.request.json = mock.CoroutineMock()
        expected_connection_record = ConnRecord(connection_id="some-id")

        with (
            mock.patch.object(
                test_module, "OutOfBandManager", autospec=True
            ) as mock_oob_mgr,
            mock.patch.object(test_module.InvitationMessage, "deserialize", mock.Mock()),
            mock.patch.object(
                test_module.web, "json_response", mock.Mock()
            ) as mock_json_response,
        ):
            mock_oob_mgr.return_value.receive_invitation = mock.CoroutineMock(
                return_value=expected_connection_record
            )

            await test_module.invitation_receive(self.request)
            mock_json_response.assert_called_once_with(
                expected_connection_record.serialize()
            )

    async def test_invitation_receive_forbidden_x(self):
        self.context.update_settings({"admin.no_receive_invites": True})
        with self.assertRaises(test_module.web.HTTPForbidden):
            await test_module.invitation_receive(self.request)

    async def test_invitation_receive_x(self):
        self.request.json = mock.CoroutineMock()

        with (
            mock.patch.object(
                test_module, "OutOfBandManager", autospec=True
            ) as mock_oob_mgr,
            mock.patch.object(test_module.InvitationMessage, "deserialize", mock.Mock()),
            mock.patch.object(test_module.web, "json_response", mock.Mock()),
        ):
            mock_oob_mgr.return_value.receive_invitation = mock.CoroutineMock(
                side_effect=test_module.StorageError("cannot write")
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.invitation_receive(self.request)

    async def test_register(self):
        mock_app = mock.MagicMock()
        mock_app.add_routes = mock.MagicMock()

        await test_module.register(mock_app)
        mock_app.add_routes.assert_called_once()

    async def test_post_process_routes(self):
        mock_app = mock.MagicMock(_state={"swagger_dict": {}})
        test_module.post_process_routes(mock_app)
        assert "tags" in mock_app._state["swagger_dict"]



from unittest import IsolatedAsyncioTestCase

from acapy_agent.tests import mock

from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......utils.testing import create_test_profile
from .. import menu_handler as handler


class TestMenuHandler(IsolatedAsyncioTestCase):
    async def test_called(self):
        request_context = RequestContext.test_context(await create_test_profile())
        request_context.connection_record = mock.MagicMock()
        request_context.connection_record.connection_id = "dummy"
        request_context.connection_ready = True

        handler.save_connection_menu = mock.CoroutineMock()
        responder = MockResponder()

        request_context.message = handler.Menu()
        handler_inst = handler.MenuHandler()
        await handler_inst.handle(request_context, responder)

        handler.save_connection_menu.assert_called_once_with(
            request_context.message,
            request_context.connection_record.connection_id,
            request_context,
        )



import json
from unittest import IsolatedAsyncioTestCase

from uuid_utils import uuid4

from ...cache.base import BaseCache
from ...cache.in_memory import InMemoryCache
from ...connections.models.conn_record import ConnRecord
from ...indy.credx.issuer import CATEGORY_REV_REG
from ...indy.issuer import IndyIssuer
from ...ledger.base import BaseLedger
from ...messaging.responder import BaseResponder
from ...protocols.issue_credential.v1_0.models.credential_exchange import (
    V10CredentialExchange,
)
from ...protocols.issue_credential.v2_0.models.cred_ex_record import V20CredExRecord
from ...revocation.models.issuer_cred_rev_record import IssuerCredRevRecord
from ...tests import mock
from ...utils.testing import create_test_profile
from .. import manager as test_module
from ..manager import RevocationManager, RevocationManagerError
from ..models.issuer_rev_reg_record import IssuerRevRegRecord

TEST_DID = "LjgpST2rjsoxYegQDRm7EL"
SCHEMA_NAME = "bc-reg"
SCHEMA_TXN = 12
SCHEMA_ID = f"{TEST_DID}:2:{SCHEMA_NAME}:1.0"
CRED_DEF_ID = f"{TEST_DID}:3:CL:12:tag1"
REV_REG_ID = f"{TEST_DID}:4:{CRED_DEF_ID}:CL_ACCUM:tag1"
TAILS_DIR = "/tmp/indy/revocation/tails_files"
TAILS_HASH = "8UW1Sz5cqoUnK9hqQk7nvtKK65t7Chu3ui866J23sFyJ"
TAILS_LOCAL = f"{TAILS_DIR}/{TAILS_HASH}"


class TestRevocationManager(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        self.manager = RevocationManager(self.profile)

    async def test_revoke_credential_publish(self):
        CRED_EX_ID = "dummy-cxid"
        CRED_REV_ID = "1"
        mock_issuer_rev_reg_record = mock.MagicMock(
            revoc_reg_id=REV_REG_ID,
            tails_local_path=TAILS_LOCAL,
            send_entry=mock.CoroutineMock(),
            clear_pending=mock.CoroutineMock(),
            pending_pub=["2"],
        )
        issuer = mock.MagicMock(IndyIssuer, autospec=True)
        issuer.revoke_credentials = mock.CoroutineMock(
            return_value=(
                json.dumps(
                    {
                        "ver": "1.0",
                        "value": {
                            "prevAccum": "1 ...",
                            "accum": "21 ...",
                            "issued": [1],
                        },
                    }
                ),
                [],
            )
        )
        self.profile.context.injector.bind_instance(IndyIssuer, issuer)

        with (
            mock.patch.object(
                test_module.IssuerCredRevRecord,
                "retrieve_by_cred_ex_id",
                mock.CoroutineMock(),
            ) as mock_retrieve,
            mock.patch.object(test_module, "IndyRevocation", autospec=True) as revoc,
            mock.patch.object(
                test_module.IssuerRevRegRecord,
                "retrieve_by_id",
                mock.CoroutineMock(return_value=mock_issuer_rev_reg_record),
            ),
        ):
            mock_retrieve.return_value = mock.MagicMock(
                rev_reg_id="dummy-rr-id", cred_rev_id=CRED_REV_ID
            )
            mock_rev_reg = mock.MagicMock(
                get_or_fetch_local_tails_path=mock.CoroutineMock()
            )
            revoc.return_value.get_issuer_rev_reg_record = mock.CoroutineMock(
                return_value=mock_issuer_rev_reg_record
            )
            revoc.return_value.get_ledger_registry = mock.CoroutineMock(
                return_value=mock_rev_reg
            )

            await self.manager.revoke_credential_by_cred_ex_id(CRED_EX_ID, publish=True)

        issuer.revoke_credentials.assert_awaited_once_with(
            mock_issuer_rev_reg_record.cred_def_id,
            mock_issuer_rev_reg_record.revoc_reg_id,
            mock_issuer_rev_reg_record.tails_local_path,
            ["2", "1"],
        )

    async def test_revoke_credential_notify(self):
        CRED_EX_ID = "dummy-cxid"
        CRED_REV_ID = "1"
        mock_issuer_rev_reg_record = mock.MagicMock(
            revoc_reg_id=REV_REG_ID,
            tails_local_path=TAILS_LOCAL,
            send_entry=mock.CoroutineMock(),
            clear_pending=mock.CoroutineMock(),
            pending_pub=["2"],
        )
        issuer = mock.MagicMock(IndyIssuer, autospec=True)
        issuer.revoke_credentials = mock.CoroutineMock(
            return_value=(
                json.dumps(
                    {
                        "ver": "1.0",
                        "value": {
                            "prevAccum": "1 ...",
                            "accum": "21 ...",
                            "issued": [1],
                        },
                    }
                ),
                [],
            )
        )
        self.profile.context.injector.bind_instance(IndyIssuer, issuer)

        with (
            mock.patch.object(
                test_module.IssuerCredRevRecord,
                "retrieve_by_cred_ex_id",
                mock.CoroutineMock(),
            ) as mock_retrieve,
            mock.patch.object(test_module, "IndyRevocation", autospec=True) as revoc,
            mock.patch.object(
                test_module.IssuerRevRegRecord,
                "retrieve_by_id",
                mock.CoroutineMock(return_value=mock_issuer_rev_reg_record),
            ),
        ):
            mock_retrieve.return_value = mock.MagicMock(
                rev_reg_id="dummy-rr-id", cred_rev_id=CRED_REV_ID
            )
            mock_rev_reg = mock.MagicMock(
                get_or_fetch_local_tails_path=mock.CoroutineMock()
            )
            revoc.return_value.get_issuer_rev_reg_record = mock.CoroutineMock(
                return_value=mock_issuer_rev_reg_record
            )
            revoc.return_value.get_ledger_registry = mock.CoroutineMock(
                return_value=mock_rev_reg
            )

            await self.manager.revoke_credential_by_cred_ex_id(
                CRED_EX_ID, publish=True, notify=True
            )

        issuer.revoke_credentials.assert_awaited_once_with(
            mock_issuer_rev_reg_record.cred_def_id,
            mock_issuer_rev_reg_record.revoc_reg_id,
            mock_issuer_rev_reg_record.tails_local_path,
            ["2", "1"],
        )

    async def test_revoke_credential_publish_endorser(self):
        conn_record = ConnRecord(
            their_label="Hello",
            their_role=ConnRecord.Role.RESPONDER.rfc160,
            alias="Bob",
        )
        session = await self.profile.session()
        await conn_record.save(session)
        await conn_record.metadata_set(
            session,
            key="endorser_info",
            value={
                "endorser_did": "test_endorser_did",
                "endorser_name": "test_endorser_name",
            },
        )
        conn_id = conn_record.connection_id
        assert conn_id is not None
        CRED_EX_ID = "dummy-cxid"
        CRED_REV_ID = "1"
        mock_issuer_rev_reg_record = mock.MagicMock(
            revoc_reg_id=REV_REG_ID,
            tails_local_path=TAILS_LOCAL,
            send_entry=mock.CoroutineMock(),
            clear_pending=mock.CoroutineMock(),
            pending_pub=["2"],
        )
        issuer = mock.MagicMock(IndyIssuer, autospec=True)
        issuer.revoke_credentials = mock.CoroutineMock(
            return_value=(
                json.dumps(
                    {
                        "ver": "1.0",
                        "value": {
                            "prevAccum": "1 ...",
                            "accum": "21 ...",
                            "issued": [1],
                        },
                    }
                ),
                [],
            )
        )
        self.profile.context.injector.bind_instance(IndyIssuer, issuer)

        with (
            mock.patch.object(
                test_module.IssuerCredRevRecord,
                "retrieve_by_cred_ex_id",
                mock.CoroutineMock(),
            ) as mock_retrieve,
            mock.patch.object(test_module, "IndyRevocation", autospec=True) as revoc,
            mock.patch.object(
                test_module.IssuerRevRegRecord,
                "retrieve_by_id",
                mock.CoroutineMock(return_value=mock_issuer_rev_reg_record),
            ),
            mock.patch.object(
                test_module.ConnRecord,
                "retrieve_by_id",
                mock.CoroutineMock(return_value=conn_record),
            ),
        ):
            mock_retrieve.return_value = mock.MagicMock(
                rev_reg_id="dummy-rr-id", cred_rev_id=CRED_REV_ID
            )
            mock_rev_reg = mock.MagicMock(
                get_or_fetch_local_tails_path=mock.CoroutineMock()
            )
            revoc.return_value.get_issuer_rev_reg_record = mock.CoroutineMock(
                return_value=mock_issuer_rev_reg_record
            )
            revoc.return_value.get_ledger_registry = mock.CoroutineMock(
                return_value=mock_rev_reg
            )

            await self.manager.revoke_credential_by_cred_ex_id(
                cred_ex_id=CRED_EX_ID,
                publish=True,
                connection_id=conn_id,
                write_ledger=False,
            )

        issuer.revoke_credentials.assert_awaited_once_with(
            mock_issuer_rev_reg_record.cred_def_id,
            mock_issuer_rev_reg_record.revoc_reg_id,
            mock_issuer_rev_reg_record.tails_local_path,
            ["2", "1"],
        )

    async def test_revoke_credential_publish_endorser_x(self):
        CRED_EX_ID = "dummy-cxid"
        CRED_REV_ID = "1"
        mock_issuer_rev_reg_record = mock.MagicMock(
            revoc_reg_id=REV_REG_ID,
            tails_local_path=TAILS_LOCAL,
            send_entry=mock.CoroutineMock(),
            clear_pending=mock.CoroutineMock(),
            pending_pub=["2"],
        )
        issuer = mock.MagicMock(IndyIssuer, autospec=True)
        issuer.revoke_credentials = mock.CoroutineMock(
            return_value=(
                json.dumps(
                    {
                        "ver": "1.0",
                        "value": {
                            "prevAccum": "1 ...",
                            "accum": "21 ...",
                            "issued": [1],
                        },
                    }
                ),
                [],
            )
        )
        self.profile.context.injector.bind_instance(IndyIssuer, issuer)

        with (
            mock.patch.object(
                test_module.IssuerCredRevRecord,
                "retrieve_by_cred_ex_id",
                mock.CoroutineMock(),
            ) as mock_retrieve,
            mock.patch.object(test_module, "IndyRevocation", autospec=True) as revoc,
            mock.patch.object(
                test_module.IssuerRevRegRecord,
                "retrieve_by_id",
                mock.CoroutineMock(return_value=mock_issuer_rev_reg_record),
            ),
            mock.patch.object(
                ConnRecord,
                "retrieve_by_id",
                mock.CoroutineMock(
                    side_effect=test_module.StorageNotFoundError("no such rec")
                ),
            ),
        ):
            mock_retrieve.return_value = mock.MagicMock(
                rev_reg_id="dummy-rr-id", cred_rev_id=CRED_REV_ID
            )
            mock_rev_reg = mock.MagicMock(
                get_or_fetch_local_tails_path=mock.CoroutineMock()
            )
            revoc.return_value.get_issuer_rev_reg_record = mock.CoroutineMock(
                return_value=mock_issuer_rev_reg_record
            )
            revoc.return_value.get_ledger_registry = mock.CoroutineMock(
                return_value=mock_rev_reg
            )
            with self.assertRaises(RevocationManagerError):
                await self.manager.revoke_credential_by_cred_ex_id(
                    cred_ex_id=CRED_EX_ID,
                    publish=True,
                    connection_id="invalid_conn_id",
                    write_ledger=False,
                )

    async def test_revoke_cred_by_cxid_not_found(self):
        CRED_EX_ID = "dummy-cxid"

        with mock.patch.object(
            test_module.IssuerCredRevRecord,
            "retrieve_by_cred_ex_id",
            mock.CoroutineMock(),
        ) as mock_retrieve:
            mock_retrieve.side_effect = test_module.StorageNotFoundError("no such rec")

            issuer = mock.MagicMock(IndyIssuer, autospec=True)
            self.profile.context.injector.bind_instance(IndyIssuer, issuer)

            with self.assertRaises(RevocationManagerError):
                await self.manager.revoke_credential_by_cred_ex_id(CRED_EX_ID)

    async def test_revoke_credential_no_rev_reg_rec(self):
        CRED_REV_ID = "1"
        V10CredentialExchange(
            credential_exchange_id="dummy-cxid",
            credential_definition_id=CRED_DEF_ID,
            role=V10CredentialExchange.ROLE_ISSUER,
            revocation_id=CRED_REV_ID,
            revoc_reg_id=REV_REG_ID,
        )

        with mock.patch.object(test_module, "IndyRevocation", autospec=True) as revoc:
            revoc.return_value.get_issuer_rev_reg_record = mock.CoroutineMock(
                return_value=None
            )

            issuer = mock.MagicMock(IndyIssuer, autospec=True)
            self.profile.context.injector.bind_instance(IndyIssuer, issuer)

            with self.assertRaises(RevocationManagerError):
                await self.manager.revoke_credential(REV_REG_ID, CRED_REV_ID)

    async def test_revoke_credential_pend(self):
        CRED_REV_ID = "1"
        mock_issuer_rev_reg_record = mock.MagicMock(mark_pending=mock.CoroutineMock())
        issuer = mock.MagicMock(IndyIssuer, autospec=True)
        self.profile.context.injector.bind_instance(IndyIssuer, issuer)

        with (
            mock.patch.object(test_module, "IndyRevocation", autospec=True) as revoc,
            mock.patch.object(
                test_module.IssuerRevRegRecord,
                "retrieve_by_id",
                mock.CoroutineMock(return_value=mock_issuer_rev_reg_record),
            ),
        ):
            revoc.return_value.get_issuer_rev_reg_record = mock.CoroutineMock(
                return_value=mock_issuer_rev_reg_record
            )

            await self.manager.revoke_credential(REV_REG_ID, CRED_REV_ID, False)

            assert (
                mock_issuer_rev_reg_record.mark_pending.call_args.args[1] == CRED_REV_ID
            )

        issuer.revoke_credentials.assert_not_awaited()

    async def test_publish_pending_revocations_endorser(self):
        deltas = [
            {
                "ver": "1.0",
                "value": {"prevAccum": "1 ...", "accum": "21 ...", "issued": [1, 2, 3]},
            },
            {
                "ver": "1.0",
                "value": {
                    "prevAccum": "21 ...",
                    "accum": "36 ...",
                    "issued": [1, 2, 3],
                },
            },
        ]

        mock_issuer_rev_reg_records = [
            mock.MagicMock(
                record_id=0,
                revoc_reg_id=REV_REG_ID,
                tails_local_path=TAILS_LOCAL,
                pending_pub=["1", "2"],
                send_entry=mock.CoroutineMock(),
                clear_pending=mock.CoroutineMock(),
            ),
            mock.MagicMock(
                record_id=1,
                revoc_reg_id=f"{TEST_DID}:4:{CRED_DEF_ID}:CL_ACCUM:tag2",
                tails_local_path=TAILS_LOCAL,
                pending_pub=["9", "99"],
                send_entry=mock.CoroutineMock(),
                clear_pending=mock.CoroutineMock(),
            ),
        ]
        conn_record = ConnRecord(
            their_label="Hello",
            their_role=ConnRecord.Role.RESPONDER.rfc160,
            alias="Bob",
        )
        session = await self.profile.session()
        await conn_record.save(session)
        await conn_record.metadata_set(
            session,
            key="endorser_info",
            value={
                "endorser_did": "test_endorser_did",
                "endorser_name": "test_endorser_name",
            },
        )
        conn_id = conn_record.connection_id
        assert conn_id is not None
        with (
            mock.patch.object(
                test_module.IssuerRevRegRecord,
                "query_by_pending",
                mock.CoroutineMock(return_value=mock_issuer_rev_reg_records),
            ),
            mock.patch.object(
                test_module.IssuerRevRegRecord,
                "retrieve_by_id",
                mock.CoroutineMock(
                    side_effect=lambda _, id, **args: mock_issuer_rev_reg_records[id]
                ),
            ),
        ):
            issuer = mock.MagicMock(IndyIssuer, autospec=True)
            issuer.merge_revocation_registry_deltas = mock.CoroutineMock(
                side_effect=deltas
            )

            issuer.revoke_credentials = mock.CoroutineMock(
                side_effect=[(json.dumps(delta), []) for delta in deltas]
            )
            self.profile.context.injector.bind_instance(IndyIssuer, issuer)
            manager = RevocationManager(self.profile)
            _, result = await manager.publish_pending_revocations(
                rrid2crid={REV_REG_ID: "2"}, connection_id=conn_id
            )
            assert result == {REV_REG_ID: ["2"]}
            mock_issuer_rev_reg_records[0].clear_pending.assert_called_once()
            mock_issuer_rev_reg_records[1].clear_pending.assert_not_called()

    async def test_publish_pending_revocations_endorser_x(self):
        deltas = [
            {
                "ver": "1.0",
                "value": {"prevAccum": "1 ...", "accum": "21 ...", "issued": [1, 2, 3]},
            },
            {
                "ver": "1.0",
                "value": {
                    "prevAccum": "21 ...",
                    "accum": "36 ...",
                    "issued": [1, 2, 3],
                },
            },
        ]

        mock_issuer_rev_reg_records = [
            mock.MagicMock(
                record_id=0,
                revoc_reg_id=REV_REG_ID,
                tails_local_path=TAILS_LOCAL,
                pending_pub=["1", "2"],
                send_entry=mock.CoroutineMock(),
                clear_pending=mock.CoroutineMock(),
            ),
            mock.MagicMock(
                record_id=1,
                revoc_reg_id=f"{TEST_DID}:4:{CRED_DEF_ID}:CL_ACCUM:tag2",
                tails_local_path=TAILS_LOCAL,
                pending_pub=["9", "99"],
                send_entry=mock.CoroutineMock(),
                clear_pending=mock.CoroutineMock(),
            ),
        ]
        with (
            mock.patch.object(
                test_module.IssuerRevRegRecord,
                "query_by_pending",
                mock.CoroutineMock(return_value=mock_issuer_rev_reg_records),
            ),
            mock.patch.object(
                test_module.IssuerRevRegRecord,
                "retrieve_by_id",
                mock.CoroutineMock(
                    side_effect=lambda _, id, **args: mock_issuer_rev_reg_records[id]
                ),
            ),
        ):
            issuer = mock.MagicMock(IndyIssuer, autospec=True)
            issuer.merge_revocation_registry_deltas = mock.CoroutineMock(
                side_effect=deltas
            )

            issuer.revoke_credentials = mock.CoroutineMock(
                side_effect=[(json.dumps(delta), []) for delta in deltas]
            )
            self.profile.context.injector.bind_instance(IndyIssuer, issuer)
            manager = RevocationManager(self.profile)
            with self.assertRaises(RevocationManagerError):
                await manager.publish_pending_revocations(
                    rrid2crid={REV_REG_ID: "2"}, connection_id="invalid_conn_id"
                )

    async def test_publish_pending_revocations_basic(self):
        deltas = [
            {
                "ver": "1.0",
                "value": {"prevAccum": "1 ...", "accum": "21 ...", "issued": [1, 2, 3]},
            },
            {
                "ver": "1.0",
                "value": {
                    "prevAccum": "21 ...",
                    "accum": "36 ...",
                    "issued": [1, 2, 3],
                },
            },
        ]

        mock_issuer_rev_reg_record = mock.MagicMock(
            revoc_reg_id=REV_REG_ID,
            tails_local_path=TAILS_LOCAL,
            pending_pub=["1", "2"],
            send_entry=mock.CoroutineMock(),
            clear_pending=mock.CoroutineMock(),
        )
        with (
            mock.patch.object(
                test_module.IssuerRevRegRecord,
                "query_by_pending",
                mock.CoroutineMock(return_value=[mock_issuer_rev_reg_record]),
            ),
            mock.patch.object(
                test_module.IssuerRevRegRecord,
                "retrieve_by_id",
                mock.CoroutineMock(return_value=mock_issuer_rev_reg_record),
            ),
        ):
            issuer = mock.MagicMock(IndyIssuer, autospec=True)
            issuer.merge_revocation_registry_deltas = mock.CoroutineMock(
                side_effect=deltas
            )

            issuer.revoke_credentials = mock.CoroutineMock(
                side_effect=[(json.dumps(delta), []) for delta in deltas]
            )
            self.profile.context.injector.bind_instance(IndyIssuer, issuer)

            _, result = await self.manager.publish_pending_revocations()
            assert result == {REV_REG_ID: ["1", "2"]}
            mock_issuer_rev_reg_record.clear_pending.assert_called_once()

    async def test_publish_pending_revocations_1_rev_reg_all(self):
        deltas = [
            {
                "ver": "1.0",
                "value": {"prevAccum": "1 ...", "accum": "21 ...", "issued": [1, 2, 3]},
            },
            {
                "ver": "1.0",
                "value": {
                    "prevAccum": "21 ...",
                    "accum": "36 ...",
                    "issued": [1, 2, 3],
                },
            },
        ]

        mock_issuer_rev_reg_records = [
            mock.MagicMock(
                record_id=0,
                revoc_reg_id=REV_REG_ID,
                tails_local_path=TAILS_LOCAL,
                pending_pub=["1", "2"],
                send_entry=mock.CoroutineMock(),
                clear_pending=mock.CoroutineMock(),
            ),
            mock.MagicMock(
                record_id=1,
                revoc_reg_id=f"{TEST_DID}:4:{CRED_DEF_ID}:CL_ACCUM:tag2",
                tails_local_path=TAILS_LOCAL,
                pending_pub=["9", "99"],
                send_entry=mock.CoroutineMock(),
                clear_pending=mock.CoroutineMock(),
            ),
        ]
        with (
            mock.patch.object(
                test_module.IssuerRevRegRecord,
                "query_by_pending",
                mock.CoroutineMock(return_value=mock_issuer_rev_reg_records),
            ),
            mock.patch.object(
                test_module.IssuerRevRegRecord,
                "retrieve_by_id",
                mock.CoroutineMock(
                    side_effect=lambda _, id, **args: mock_issuer_rev_reg_records[id]
                ),
            ),
        ):
            issuer = mock.MagicMock(IndyIssuer, autospec=True)
            issuer.merge_revocation_registry_deltas = mock.CoroutineMock(
                side_effect=deltas
            )

            issuer.revoke_credentials = mock.CoroutineMock(
                side_effect=[(json.dumps(delta), []) for delta in deltas]
            )
            self.profile.context.injector.bind_instance(IndyIssuer, issuer)

            _, result = await self.manager.publish_pending_revocations({REV_REG_ID: None})
            assert result == {REV_REG_ID: ["1", "2"]}
            mock_issuer_rev_reg_records[0].clear_pending.assert_called_once()
            mock_issuer_rev_reg_records[1].clear_pending.assert_not_called()

    async def test_publish_pending_revocations_1_rev_reg_some(self):
        deltas = [
            {
                "ver": "1.0",
                "value": {"prevAccum": "1 ...", "accum": "21 ...", "issued": [1, 2, 3]},
            },
            {
                "ver": "1.0",
                "value": {
                    "prevAccum": "21 ...",
                    "accum": "36 ...",
                    "issued": [1, 2, 3],
                },
            },
        ]

        mock_issuer_rev_reg_records = [
            mock.MagicMock(
                record_id=0,
                revoc_reg_id=REV_REG_ID,
                tails_local_path=TAILS_LOCAL,
                pending_pub=["1", "2"],
                send_entry=mock.CoroutineMock(),
                clear_pending=mock.CoroutineMock(),
            ),
            mock.MagicMock(
                record_id=1,
                revoc_reg_id=f"{TEST_DID}:4:{CRED_DEF_ID}:CL_ACCUM:tag2",
                tails_local_path=TAILS_LOCAL,
                pending_pub=["9", "99"],
                send_entry=mock.CoroutineMock(),
                clear_pending=mock.CoroutineMock(),
            ),
        ]
        with (
            mock.patch.object(
                test_module.IssuerRevRegRecord,
                "query_by_pending",
                mock.CoroutineMock(return_value=mock_issuer_rev_reg_records),
            ),
            mock.patch.object(
                test_module.IssuerRevRegRecord,
                "retrieve_by_id",
                mock.CoroutineMock(
                    side_effect=lambda _, id, **args: mock_issuer_rev_reg_records[id]
                ),
            ),
        ):
            issuer = mock.MagicMock(IndyIssuer, autospec=True)
            issuer.merge_revocation_registry_deltas = mock.CoroutineMock(
                side_effect=deltas
            )

            issuer.revoke_credentials = mock.CoroutineMock(
                side_effect=[(json.dumps(delta), []) for delta in deltas]
            )
            self.profile.context.injector.bind_instance(IndyIssuer, issuer)

            _, result = await self.manager.publish_pending_revocations({REV_REG_ID: "2"})
            assert result == {REV_REG_ID: ["2"]}
            mock_issuer_rev_reg_records[0].clear_pending.assert_called_once()
            mock_issuer_rev_reg_records[1].clear_pending.assert_not_called()

    async def test_clear_pending(self):
        REV_REG_ID_2 = f"{TEST_DID}:4:{CRED_DEF_ID}:CL_ACCUM:tag2"
        mock_issuer_rev_reg_records = [
            test_module.IssuerRevRegRecord(
                revoc_reg_id=REV_REG_ID,
                tails_local_path=TAILS_LOCAL,
                pending_pub=["1", "2"],
            ),
            test_module.IssuerRevRegRecord(
                revoc_reg_id=REV_REG_ID_2,
                tails_local_path=TAILS_LOCAL,
                pending_pub=["9", "99"],
            ),
        ]
        with mock.patch.object(
            test_module.IssuerRevRegRecord,
            "query_by_pending",
            mock.CoroutineMock(return_value=mock_issuer_rev_reg_records),
        ):
            result = await self.manager.clear_pending_revocations()
            assert result[REV_REG_ID] == []
            assert result[REV_REG_ID_2] == []

    async def test_clear_pending_1_rev_reg_all(self):
        REV_REG_ID_2 = f"{TEST_DID}:4:{CRED_DEF_ID}:CL_ACCUM:tag2"

        mock_issuer_rev_reg_records = [
            test_module.IssuerRevRegRecord(
                revoc_reg_id=REV_REG_ID,
                tails_local_path=TAILS_LOCAL,
                pending_pub=["1", "2"],
            ),
            test_module.IssuerRevRegRecord(
                revoc_reg_id=REV_REG_ID_2,
                tails_local_path=TAILS_LOCAL,
                pending_pub=["9", "99"],
            ),
        ]
        with mock.patch.object(
            test_module.IssuerRevRegRecord,
            "query_by_pending",
            mock.CoroutineMock(return_value=mock_issuer_rev_reg_records),
        ):
            result = await self.manager.clear_pending_revocations({REV_REG_ID: []})
            assert result[REV_REG_ID] == []
            assert result.get(REV_REG_ID_2) is None

    async def test_clear_pending_1_rev_reg_some(self):
        REV_REG_ID_2 = f"{TEST_DID}:4:{CRED_DEF_ID}:CL_ACCUM:tag2"
        mock_issuer_rev_reg_records = [
            test_module.IssuerRevRegRecord(
                revoc_reg_id=REV_REG_ID,
                tails_local_path=TAILS_LOCAL,
                pending_pub=["1", "2"],
            ),
            test_module.IssuerRevRegRecord(
                revoc_reg_id=REV_REG_ID_2,
                tails_local_path=TAILS_LOCAL,
                pending_pub=["99"],
            ),
        ]
        with mock.patch.object(
            test_module.IssuerRevRegRecord,
            "query_by_pending",
            mock.CoroutineMock(return_value=mock_issuer_rev_reg_records),
        ):
            result = await self.manager.clear_pending_revocations({REV_REG_ID: ["9"]})

            assert result[REV_REG_ID] == ["1", "2"]
            assert result.get(REV_REG_ID_2) is None

    async def test_clear_pending_both(self):
        REV_REG_ID_2 = f"{TEST_DID}:4:{CRED_DEF_ID}:CL_ACCUM:tag2"
        mock_issuer_rev_reg_records = [
            test_module.IssuerRevRegRecord(
                revoc_reg_id=REV_REG_ID,
                tails_local_path=TAILS_LOCAL,
                pending_pub=["1", "2"],
            ),
            test_module.IssuerRevRegRecord(
                revoc_reg_id=REV_REG_ID_2,
                tails_local_path=TAILS_LOCAL,
                pending_pub=["99"],
            ),
        ]
        with mock.patch.object(
            test_module.IssuerRevRegRecord,
            "query_by_pending",
            mock.CoroutineMock(return_value=mock_issuer_rev_reg_records),
        ):
            result = await self.manager.clear_pending_revocations(
                {REV_REG_ID: ["1"], REV_REG_ID_2: ["99"]}
            )
            assert result[REV_REG_ID] == ["2"]
            assert result[REV_REG_ID_2] == []

    async def test_retrieve_records(self):
        session = await self.profile.session()
        for index in range(2):
            exchange_record = V10CredentialExchange(
                connection_id=str(index),
                thread_id=str(1000 + index),
                initiator=V10CredentialExchange.INITIATOR_SELF,
                role=V10CredentialExchange.ROLE_ISSUER,
            )
            await exchange_record.save(session)

        for _ in range(2):  # second pass gets from cache
            for index in range(2):
                ret_ex = await V10CredentialExchange.retrieve_by_connection_and_thread(
                    session, str(index), str(1000 + index)
                )
                assert ret_ex.connection_id == str(index)
                assert ret_ex.thread_id == str(1000 + index)

    async def test_set_revoked_state_v1(self):
        CRED_REV_ID = "1"

        async with self.profile.session() as session:
            exchange_record = V10CredentialExchange(
                connection_id="mark-revoked-cid",
                thread_id="mark-revoked-tid",
                initiator=V10CredentialExchange.INITIATOR_SELF,
                revoc_reg_id=REV_REG_ID,
                revocation_id=CRED_REV_ID,
                role=V10CredentialExchange.ROLE_ISSUER,
                state=V10CredentialExchange.STATE_ISSUED,
            )
            await exchange_record.save(session)

            crev_record = IssuerCredRevRecord(
                cred_ex_id=exchange_record.credential_exchange_id,
                cred_def_id=CRED_DEF_ID,
                rev_reg_id=REV_REG_ID,
                cred_rev_id=CRED_REV_ID,
                state=IssuerCredRevRecord.STATE_ISSUED,
            )
            await crev_record.save(session)

        await self.manager.set_cred_revoked_state(REV_REG_ID, [CRED_REV_ID])

        async with self.profile.session() as session:
            check_exchange_record = await V10CredentialExchange.retrieve_by_id(
                session, exchange_record.credential_exchange_id
            )
            assert (
                check_exchange_record.state
                == V10CredentialExchange.STATE_CREDENTIAL_REVOKED
            )

            check_crev_record = await IssuerCredRevRecord.retrieve_by_id(
                session, crev_record.record_id
            )
            assert check_crev_record.state == IssuerCredRevRecord.STATE_REVOKED

    async def test_set_revoked_state_v2(self):
        CRED_REV_ID = "1"

        async with self.profile.session() as session:
            exchange_record = V20CredExRecord(
                connection_id="mark-revoked-cid",
                thread_id="mark-revoked-tid",
                initiator=V20CredExRecord.INITIATOR_SELF,
                role=V20CredExRecord.ROLE_ISSUER,
                state=V20CredExRecord.STATE_ISSUED,
            )
            await exchange_record.save(session)

            crev_record = IssuerCredRevRecord(
                cred_ex_id=exchange_record.cred_ex_id,
                cred_def_id=CRED_DEF_ID,
                rev_reg_id=REV_REG_ID,
                cred_rev_id=CRED_REV_ID,
                state=IssuerCredRevRecord.STATE_ISSUED,
            )
            await crev_record.save(session)

        await self.manager.set_cred_revoked_state(REV_REG_ID, [CRED_REV_ID])

        async with self.profile.session() as session:
            check_exchange_record = await V20CredExRecord.retrieve_by_id(
                session, exchange_record.cred_ex_id
            )
            assert check_exchange_record.state == V20CredExRecord.STATE_CREDENTIAL_REVOKED

            check_crev_record = await IssuerCredRevRecord.retrieve_by_id(
                session, crev_record.record_id
            )
            assert check_crev_record.state == IssuerCredRevRecord.STATE_REVOKED

    @mock.patch.object(
        ConnRecord,
        "retrieve_by_id",
        mock.CoroutineMock(
            return_value=mock.MagicMock(
                connection_id="endorser-id",
                metadata_get=mock.CoroutineMock(
                    return_value={"endorser_did": "test_endorser_did"}
                ),
            )
        ),
    )
    @mock.patch.object(
        IssuerRevRegRecord,
        "fix_ledger_entry",
        mock.CoroutineMock(
            return_value=(
                "1 ...",
                {
                    "ver": "1.0",
                    "value": {
                        "prevAccum": "1 ...",
                        "accum": "fixed-accum",
                        "issued": [1],
                    },
                },
                [1],
            )
        ),
    )
    async def test_fix_and_publish_from_invalid_accum_err(
        self,
    ):
        # Setup
        self.profile.context.injector.bind_instance(BaseCache, InMemoryCache())
        self.profile.context.injector.bind_instance(
            BaseResponder, mock.MagicMock(BaseResponder, autospec=True)
        )
        mock_ledger = mock.MagicMock(BaseLedger, autospec=True)
        mock_ledger.get_revoc_reg_delta = mock.CoroutineMock(
            side_effect=[
                ({"value": {"accum": "other-accum"}}, None),
                ({"value": {"accum": "invalid-accum"}}, None),
            ]
        )
        mock_ledger.send_revoc_reg_entry = mock.CoroutineMock(
            return_value={"result": {"txn": "..."}}
        )
        self.profile.context.injector.bind_instance(BaseLedger, mock_ledger)
        self.profile.context.settings.set_value(
            "ledger.genesis_transactions", {"txn": "..."}
        )
        self.profile.context.settings.set_value("endorser.endorser_alias", "endorser")

        async with self.profile.session() as session:
            # Add an endorser connection
            await session.handle.insert(
                ConnRecord.RECORD_TYPE,
                name="endorser",
                value_json={"connection_id": "endorser", "alias": "endorser"},
            )
            record = ConnRecord(
                alias="endorser",
            )
            await record.save(session)

            # Add a couple rev reg records
            for _ in range(2):
                await session.handle.insert(
                    IssuerRevRegRecord.RECORD_TYPE,
                    name=str(uuid4()),
                    value_json={
                        "revoc_reg_id": "test-rr-id",
                    },
                )

            # Need a matching revocation_reg record
            await session.handle.insert(
                CATEGORY_REV_REG,
                name="test-rr-id",
                value_json={
                    "value": {
                        "accum": "invalid-accum",
                        "revoked": [1],
                    }
                },
            )

        # Execute
        await self.manager.fix_and_publish_from_invalid_accum_err("invalid-accum")



DID_zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq = {
    "@context": ["https://www.w3.org/ns/did/v1", "https://w3id.org/security/multikey/v1"],
    "id": "did:key:zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq",
    "verificationMethod": [
        {
            "id": "did:key:zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq#zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq",
            "type": "Multikey",
            "controller": "did:key:zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq",
            "publicKeyMultibase": "zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq",
        }
    ],
    "authentication": [
        "did:key:zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq#zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq"
    ],
    "assertionMethod": [
        "did:key:zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq#zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq"
    ],
    "capabilityDelegation": [
        "did:key:zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq#zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq"
    ],
    "capabilityInvocation": [
        "did:key:zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq#zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq"
    ],
    "keyAgreement": [],
}



"""Test settings routes."""

# pylint: disable=redefined-outer-name

import pytest
import pytest_asyncio

from ...admin.request_context import AdminRequestContext
from ...multitenant.base import BaseMultitenantManager
from ...multitenant.manager import MultitenantManager
from ...tests import mock
from ...utils.testing import create_test_profile
from .. import routes as test_module


@pytest.fixture
def mock_response():
    json_response = mock.MagicMock()
    temp_value = test_module.web.json_response
    test_module.web.json_response = json_response
    yield json_response
    test_module.web.json_response = temp_value


@pytest_asyncio.fixture
async def profile():
    profile = await create_test_profile()
    yield profile


@pytest_asyncio.fixture
async def admin_profile():
    profile = await create_test_profile(
        settings={
            "admin.admin_api_key": "secret-key",
        }
    )
    yield profile


@pytest.mark.asyncio
async def test_get_profile_settings(mock_response, admin_profile, profile):
    admin_profile.settings.update(
        {
            "admin.admin_client_max_request_size": 1,
            "debug.auto_respond_credential_offer": True,
            "debug.auto_respond_credential_request": True,
            "debug.auto_respond_presentation_proposal": True,
            "debug.auto_verify_presentation": True,
            "debug.auto_accept_invites": True,
            "debug.auto_accept_requests": True,
        }
    )
    request_dict = {
        "context": AdminRequestContext(
            profile=admin_profile,
        ),
    }
    request = mock.MagicMock(
        query={},
        json=mock.CoroutineMock(return_value={}),
        __getitem__=lambda _, k: request_dict[k],
        headers={"x-api-key": "secret-key"},
    )
    await test_module.get_profile_settings(request)
    assert mock_response.call_args[0][0] == {
        "debug.auto_respond_credential_offer": True,
        "debug.auto_respond_credential_request": True,
        "debug.auto_verify_presentation": True,
        "debug.auto_accept_invites": True,
        "debug.auto_accept_requests": True,
    }
    # Multitenant
    multi_tenant_manager = MultitenantManager(profile)
    profile.context.injector.bind_instance(
        BaseMultitenantManager,
        multi_tenant_manager,
    )
    request_dict = {
        "context": AdminRequestContext(
            profile=profile,
            root_profile=profile,
            metadata={
                "wallet_id": "walletid",
                "wallet_key": "walletkey",
            },
        ),
    }
    request = mock.MagicMock(
        query={},
        json=mock.CoroutineMock(return_value={}),
        __getitem__=lambda _, k: request_dict[k],
    )
    with mock.patch.object(
        multi_tenant_manager, "get_wallet_and_profile"
    ) as get_wallet_and_profile:
        get_wallet_and_profile.return_value = (
            mock.MagicMock(
                settings={
                    "admin.admin_client_max_request_size": 1,
                    "debug.auto_respond_credential_offer": True,
                    "debug.auto_respond_credential_request": True,
                    "debug.auto_respond_presentation_proposal": True,
                    "debug.auto_verify_presentation": True,
                    "debug.auto_accept_invites": True,
                    "debug.auto_accept_requests": True,
                }
            ),
            profile,
        )
        await test_module.get_profile_settings(request)
    assert mock_response.call_args[0][0] == {
        "debug.auto_respond_credential_offer": True,
        "debug.auto_respond_credential_request": True,
        "debug.auto_verify_presentation": True,
        "debug.auto_accept_invites": True,
        "debug.auto_accept_requests": True,
        "wallet.type": "askar",
    }


@pytest.mark.asyncio
async def test_update_profile_settings(mock_response, profile):
    profile.settings.update(
        {
            "public_invites": True,
            "debug.invite_public": True,
            "debug.auto_accept_invites": True,
            "debug.auto_accept_requests": True,
            "auto_ping_connection": True,
        }
    )
    request_dict = {
        "context": AdminRequestContext(
            profile=profile,
        ),
    }
    request = mock.MagicMock(
        query={},
        json=mock.CoroutineMock(
            return_value={
                "extra_settings": {
                    "ACAPY_INVITE_PUBLIC": False,
                    "ACAPY_PUBLIC_INVITES": False,
                    "ACAPY_AUTO_ACCEPT_INVITES": False,
                    "ACAPY_AUTO_ACCEPT_REQUESTS": False,
                    "ACAPY_AUTO_PING_CONNECTION": False,
                }
            }
        ),
        __getitem__=lambda _, k: request_dict[k],
    )
    await test_module.update_profile_settings(request)
    assert mock_response.call_args[0][0] == {
        "public_invites": False,
        "debug.invite_public": False,
        "debug.auto_accept_invites": False,
        "debug.auto_accept_requests": False,
        "auto_ping_connection": False,
        "wallet.type": "askar",
    }
    # Multitenant
    multi_tenant_manager = MultitenantManager(profile)
    profile.context.injector.bind_instance(
        BaseMultitenantManager,
        multi_tenant_manager,
    )

    request_dict = {
        "context": AdminRequestContext(
            profile=profile,
            root_profile=profile,
            metadata={
                "wallet_id": "walletid",
                "wallet_key": "walletkey",
            },
        ),
    }
    request = mock.MagicMock(
        query={},
        json=mock.CoroutineMock(
            return_value={
                "extra_settings": {
                    "ACAPY_INVITE_PUBLIC": False,
                    "ACAPY_PUBLIC_INVITES": False,
                    "ACAPY_AUTO_ACCEPT_INVITES": False,
                    "ACAPY_AUTO_ACCEPT_REQUESTS": False,
                    "ACAPY_AUTO_PING_CONNECTION": False,
                }
            }
        ),
        __getitem__=lambda _, k: request_dict[k],
    )
    with (
        mock.patch.object(multi_tenant_manager, "update_wallet") as update_wallet,
        mock.patch.object(
            multi_tenant_manager, "get_wallet_and_profile"
        ) as get_wallet_and_profile,
    ):
        get_wallet_and_profile.return_value = (
            mock.MagicMock(
                settings={
                    "admin.admin_client_max_request_size": 1,
                    "debug.auto_respond_credential_offer": True,
                    "debug.auto_respond_credential_request": True,
                    "debug.auto_respond_presentation_proposal": True,
                    "debug.auto_verify_presentation": True,
                    "public_invites": False,
                    "debug.invite_public": False,
                    "debug.auto_accept_invites": False,
                    "debug.auto_accept_requests": False,
                    "auto_ping_connection": False,
                }
            ),
            profile,
        )
        update_wallet.return_value = mock.MagicMock(
            settings={
                "public_invites": False,
                "debug.invite_public": False,
                "debug.auto_accept_invites": False,
                "debug.auto_accept_requests": False,
                "auto_ping_connection": False,
            }
        )
        await test_module.update_profile_settings(request)
    assert mock_response.call_args[0][0] == {
        "public_invites": False,
        "debug.invite_public": False,
        "debug.auto_accept_invites": False,
        "debug.auto_accept_requests": False,
        "auto_ping_connection": False,
        "debug.auto_respond_credential_offer": True,
        "debug.auto_respond_credential_request": True,
        "debug.auto_verify_presentation": True,
        "wallet.type": "askar",
    }



from ...core.profile import Profile
from ..cache import ProfileCache


class MockProfile(Profile):
    def session(self, context=None): ...

    def transaction(self, context=None): ...


def test_get_not_in_cache():
    cache = ProfileCache(1)

    assert cache.get("1") is None


def test_put_get_in_cache():
    cache = ProfileCache(1)

    profile = MockProfile()
    cache.put("1", profile)

    assert cache.get("1") is profile


def test_remove():
    cache = ProfileCache(1)

    profile = MockProfile()
    cache.put("1", profile)

    assert cache.get("1") is profile

    cache.remove("1")

    assert cache.get("1") is None


def test_has_true():
    cache = ProfileCache(1)

    profile = MockProfile()

    assert cache.has("1") is False
    cache.put("1", profile)
    assert cache.has("1") is True


def test_cleanup():
    cache = ProfileCache(1)

    cache.put("1", MockProfile())

    assert len(cache.profiles) == 1

    cache.put("2", MockProfile())

    assert len(cache.profiles) == 1
    assert cache.get("1") is None


def test_cleanup_lru():
    cache = ProfileCache(3)

    cache.put("1", MockProfile())
    cache.put("2", MockProfile())
    cache.put("3", MockProfile())

    assert len(cache.profiles) == 3

    cache.get("1")

    cache.put("4", MockProfile())

    assert len(cache._cache) == 3
    assert cache.get("1")
    assert cache.get("2") is None
    assert cache.get("3")
    assert cache.get("4")


def test_rescue_open_profile():
    cache = ProfileCache(3)

    cache.put("1", MockProfile())
    cache.put("2", MockProfile())
    cache.put("3", MockProfile())

    assert len(cache.profiles) == 3

    held = cache.profiles["1"]
    cache.put("4", MockProfile())

    assert len(cache.profiles) == 4
    assert len(cache._cache) == 3

    cache.get("1")

    assert len(cache.profiles) == 3
    assert len(cache._cache) == 3
    assert cache.get("1")
    assert cache.get("2") is None
    assert cache.get("3")
    assert cache.get("4")



from unittest import IsolatedAsyncioTestCase

from ......messaging.decorators.attach_decorator import AttachDecorator
from ......messaging.models.base import BaseModelError
from .....didcomm_prefix import DIDCommPrefix
from ...message_types import ATTACHMENT_FORMAT, CRED_20_PREVIEW, CRED_20_PROPOSAL
from ..cred_format import V20CredFormat
from ..cred_proposal import V20CredProposal
from ..inner.cred_preview import V20CredAttrSpec, V20CredPreview

TEST_PREVIEW = V20CredPreview(
    attributes=(
        V20CredAttrSpec.list_plain({"test": "123", "hello": "world"})
        + [V20CredAttrSpec(name="icon", value="cG90YXRv", mime_type="image/png")]
    )
)

TEST_INDY_FILTER = {
    "schema_id": "GMm4vMw8LLrLJjp81kRRLp:2:ahoy:1560364003.0",
    "cred_def_id": "GMm4vMw8LLrLJjp81kRRLp:3:CL:12:tag",
}


class TestV20CredProposal(IsolatedAsyncioTestCase):
    """Credential proposal tests."""

    async def test_init(self):
        """Test initializer."""
        cred_proposal = V20CredProposal(
            comment="Hello World",
            credential_preview=TEST_PREVIEW,
            formats=[
                V20CredFormat(
                    attach_id="indy",
                    format_=ATTACHMENT_FORMAT[CRED_20_PROPOSAL][
                        V20CredFormat.Format.INDY.api
                    ],
                )
            ],
            filters_attach=[AttachDecorator.data_base64(TEST_INDY_FILTER, ident="indy")],
        )
        assert cred_proposal.credential_preview == TEST_PREVIEW
        assert cred_proposal.attachment() == TEST_INDY_FILTER
        assert cred_proposal._type == DIDCommPrefix.qualify_current(CRED_20_PROPOSAL)

    async def test_attachment_no_target_format(self):
        """Test attachment behaviour for only unknown formats."""

        x_cred_proposal = V20CredProposal(
            comment="Test",
            formats=[V20CredFormat(attach_id="not_indy", format_="not_indy")],
            filters_attach=[
                AttachDecorator.data_base64(
                    ident="not_indy",
                    mapping=TEST_PREVIEW.serialize(),
                )
            ],
        )
        assert x_cred_proposal.attachment() is None

    async def test_deserialize(self):
        """Test deserialization."""
        obj = {
            "@type": "https://didcomm.org/issue-credential/2.0/propose-credential",
            "@id": "56dfd607-e03b-4175-8e36-c49329da891b",
            "comment": "Hello World",
            "credential_preview": {
                "@type": "https://didcomm.org/issue-credential/2.0/credential-preview",
                "attributes": [
                    {"name": "test", "value": "123"},
                    {"name": "hello", "value": "world"},
                    {"name": "icon", "mime-type": "image/png", "value": "cG90YXRv"},
                ],
            },
            "formats": [
                {
                    "attach_id": "indy",
                    "format": ATTACHMENT_FORMAT[CRED_20_PROPOSAL][
                        V20CredFormat.Format.INDY.api
                    ],
                }
            ],
            "filters~attach": [
                {
                    "@id": "indy",
                    "mime-type": "application/json",
                    "data": {
                        "base64": (
                            "eyJzY2hlbWFfaWQiOiAiR01tNHZNdzhMTHJMSmpw"
                            "ODFrUlJMcDoyOmFob3k6MTU2MDM2NDAwMy4wIiwg"
                            "ImNyZWRfZGVmX2lkIjogIkdNbTR2TXc4TExyTEpq"
                            "cDgxa1JSTHA6MzpDTDoxMjp0YWcifQ=="
                        )
                    },
                }
            ],
        }
        cred_proposal = V20CredProposal.deserialize(obj)
        assert type(cred_proposal) is V20CredProposal

        obj["filters~attach"][0]["data"]["base64"] = "eyJub3QiOiAiaW5keSJ9"  # not indy
        with self.assertRaises(BaseModelError):
            V20CredProposal.deserialize(obj)

        obj["filters~attach"][0]["@id"] = "xxx"
        with self.assertRaises(BaseModelError):
            V20CredProposal.deserialize(obj)

        obj["filters~attach"].append(  # more attachments than formats
            {
                "@id": "def",
                "mime-type": "application/json",
                "data": {"base64": "eyJub3QiOiAiaW5keSJ9"},
            }
        )
        with self.assertRaises(BaseModelError):
            V20CredProposal.deserialize(obj)

        cred_proposal.formats.append(  # unknown format: no validation
            V20CredFormat(
                attach_id="not_indy",
                format_="not_indy",
            )
        )
        obj = cred_proposal.serialize()
        obj["filters~attach"].append(
            {
                "@id": "not_indy",
                "mime-type": "application/json",
                "data": {"base64": "eyJub3QiOiAiaW5keSJ9"},
            }
        )
        V20CredProposal.deserialize(obj)

    async def test_serialize(self):
        """Test serialization."""

        cred_proposal = V20CredProposal(
            comment="Hello World",
            credential_preview=TEST_PREVIEW,
            formats=[
                V20CredFormat(
                    attach_id="indy",
                    format_=ATTACHMENT_FORMAT[CRED_20_PROPOSAL][
                        V20CredFormat.Format.INDY.api
                    ],
                )
            ],
            filters_attach=[AttachDecorator.data_base64(TEST_INDY_FILTER, ident="indy")],
        )

        cred_proposal_dict = cred_proposal.serialize()
        cred_proposal_dict.pop("@id")

        assert cred_proposal_dict == {
            "@type": DIDCommPrefix.qualify_current(CRED_20_PROPOSAL),
            "comment": "Hello World",
            "credential_preview": {
                "@type": DIDCommPrefix.qualify_current(CRED_20_PREVIEW),
                "attributes": [
                    {"name": "test", "value": "123"},
                    {"name": "hello", "value": "world"},
                    {"name": "icon", "mime-type": "image/png", "value": "cG90YXRv"},
                ],
            },
            "formats": [
                {
                    "attach_id": "indy",
                    "format": ATTACHMENT_FORMAT[CRED_20_PROPOSAL][
                        V20CredFormat.Format.INDY.api
                    ],
                }
            ],
            "filters~attach": [
                {
                    "@id": "indy",
                    "mime-type": "application/json",
                    "data": {
                        "base64": (
                            "eyJzY2hlbWFfaWQiOiAiR01tNHZNdzhMTHJMSmpw"
                            "ODFrUlJMcDoyOmFob3k6MTU2MDM2NDAwMy4wIiwg"
                            "ImNyZWRfZGVmX2lkIjogIkdNbTR2TXc4TExyTEpq"
                            "cDgxa1JSTHA6MzpDTDoxMjp0YWcifQ=="
                        )
                    },
                }
            ],
        }

    async def test_serialize_minimal(self):
        """Test serialization."""

        cred_proposal = V20CredProposal(
            credential_preview=None,
            formats=[
                V20CredFormat(
                    attach_id="indy",
                    format_=ATTACHMENT_FORMAT[CRED_20_PROPOSAL][
                        V20CredFormat.Format.INDY.api
                    ],
                )
            ],
            filters_attach=[AttachDecorator.data_base64({}, ident="indy")],
        )

        cred_proposal_dict = cred_proposal.serialize()
        cred_proposal_dict.pop("@id")

        assert cred_proposal_dict == {
            "@type": DIDCommPrefix.qualify_current(CRED_20_PROPOSAL),
            "filters~attach": [
                {
                    "@id": "indy",
                    "mime-type": "application/json",
                    "data": {"base64": "e30="},
                }
            ],
            "formats": [
                {
                    "attach_id": "indy",
                    "format": ATTACHMENT_FORMAT[CRED_20_PROPOSAL][
                        V20CredFormat.Format.INDY.api
                    ],
                }
            ],
        }


class TestV20CredProposalSchema(IsolatedAsyncioTestCase):
    """Test credential proposal schema."""

    async def test_make_model(self):
        """Test making model."""
        cred_proposal = V20CredProposal(
            credential_preview=TEST_PREVIEW,
            formats=[
                V20CredFormat(
                    attach_id="indy",
                    format_=ATTACHMENT_FORMAT[CRED_20_PROPOSAL][
                        V20CredFormat.Format.INDY.api
                    ],
                )
            ],
            filters_attach=[AttachDecorator.data_base64(TEST_INDY_FILTER, ident="indy")],
        )

        data = cred_proposal.serialize()
        model_instance = V20CredProposal.deserialize(data)
        assert isinstance(model_instance, V20CredProposal)



"""Tooling to validate DID creation parameters."""

import hashlib
from typing import Optional

from acapy_agent.did.did_key import DIDKey
from acapy_agent.wallet.did_method import (
    INDY,
    KEY,
    SOV,
    DIDMethod,
    DIDMethods,
    HolderDefinedDid,
)
from acapy_agent.wallet.error import WalletError
from acapy_agent.wallet.key_type import KeyType
from acapy_agent.wallet.util import bytes_to_b58


class DIDParametersValidation:
    """A utility class to check compatibility of provided DID creation parameters."""

    def __init__(self, did_methods: DIDMethods):
        """:param did_methods: DID method registry relevant for the validation."""
        self.did_methods = did_methods

    @staticmethod
    def validate_key_type(method: DIDMethod, key_type: KeyType):
        """Validate compatibility of the DID method with the desired key type."""
        # validate key_type
        if not method.supports_key_type(key_type):
            raise WalletError(
                f"Invalid key type {key_type.key_type}"
                f" for DID method {method.method_name}"
            )

    def validate_or_derive_did(
        self,
        method: DIDMethod,
        key_type: KeyType,
        verkey: bytes,
        did: Optional[str],
    ) -> str:
        """Validate compatibility of the provided did (if any) with the given DID method.

        If no DID was provided, automatically derive one for methods that support it.
        """
        if method.holder_defined_did() == HolderDefinedDid.NO and did:
            raise WalletError(
                f"Not allowed to set DID for DID method '{method.method_name}'"
            )
        elif method.holder_defined_did() == HolderDefinedDid.REQUIRED and not did:
            raise WalletError(f"Providing a DID is required {method.method_name}")
        elif not self.did_methods.registered(method.method_name):
            raise WalletError(
                f"Unsupported DID method for current storage: {method.method_name}"
            )

        # We need some did method specific handling. If more did methods
        # are added it is probably better create a did method specific handler
        elif method == KEY:
            return DIDKey.from_public_key(verkey, key_type).did
        elif method == SOV:
            return bytes_to_b58(verkey[:16]) if not did else did
        elif method == INDY:
            return bytes_to_b58(hashlib.sha256(verkey).digest()[:16]) if not did else did

        return did



DID_V1 = {
    "@context": {
        "@protected": True,
        "id": "@id",
        "type": "@type",
        "alsoKnownAs": {
            "@id": "https://www.w3.org/ns/activitystreams#alsoKnownAs",
            "@type": "@id",
        },
        "assertionMethod": {
            "@id": "https://w3id.org/security#assertionMethod",
            "@type": "@id",
            "@container": "@set",
        },
        "authentication": {
            "@id": "https://w3id.org/security#authenticationMethod",
            "@type": "@id",
            "@container": "@set",
        },
        "capabilityDelegation": {
            "@id": "https://w3id.org/security#capabilityDelegationMethod",
            "@type": "@id",
            "@container": "@set",
        },
        "capabilityInvocation": {
            "@id": "https://w3id.org/security#capabilityInvocationMethod",
            "@type": "@id",
            "@container": "@set",
        },
        "controller": {"@id": "https://w3id.org/security#controller", "@type": "@id"},
        "keyAgreement": {
            "@id": "https://w3id.org/security#keyAgreementMethod",
            "@type": "@id",
            "@container": "@set",
        },
        "service": {
            "@id": "https://www.w3.org/ns/did#service",
            "@type": "@id",
            "@context": {
                "@protected": True,
                "id": "@id",
                "type": "@type",
                "serviceEndpoint": {
                    "@id": "https://www.w3.org/ns/did#serviceEndpoint",
                    "@type": "@id",
                },
            },
        },
        "verificationMethod": {
            "@id": "https://w3id.org/security#verificationMethod",
            "@type": "@id",
        },
    }
}



"""Record used to represent individual menu options in an action menu."""

from typing import Optional

from marshmallow import EXCLUDE, fields

from .....messaging.models.base import BaseModel, BaseModelSchema
from .menu_form import MenuForm, MenuFormSchema


class MenuOption(BaseModel):
    """Instance of a menu option associated with an action menu."""

    class Meta:
        """Menu option metadata."""

        schema_class = "MenuOptionSchema"

    def __init__(
        self,
        *,
        name: Optional[str] = None,
        title: Optional[str] = None,
        description: Optional[str] = None,
        disabled: Optional[bool] = None,
        form: Optional[MenuForm] = None,
    ):
        """Initialize a MenuOption instance.

        Args:
            name: The menu option name (unique ID)
            title: The menu option title
            description: Additional descriptive text for the menu option
            disabled: If the option should be shown as disabled
            form: A form to display when the option is selected
        """
        self.name = name
        self.title = title
        self.description = description
        self.disabled = disabled
        self.form = form


class MenuOptionSchema(BaseModelSchema):
    """MenuOption schema."""

    class Meta:
        """MenuOptionSchema metadata."""

        model_class = MenuOption
        unknown = EXCLUDE

    name = fields.Str(
        required=True,
        metadata={
            "description": "Menu option name (unique identifier)",
            "example": "window_prefs",
        },
    )
    title = fields.Str(
        required=True,
        metadata={"description": "Menu option title", "example": "Window Preferences"},
    )
    description = fields.Str(
        required=False,
        metadata={
            "description": "Additional descriptive text for menu option",
            "example": "Window display preferences",
        },
    )
    disabled = fields.Bool(
        required=False,
        metadata={
            "description": "Whether to show option as disabled",
            "example": "False",
        },
    )
    form = fields.Nested(MenuFormSchema(), required=False)



"""Credential definition utilities."""

import re

from marshmallow import fields

from ...core.profile import Profile
from ..models.openapi import OpenAPISchema
from ..valid import (
    INDY_CRED_DEF_ID_EXAMPLE,
    INDY_CRED_DEF_ID_VALIDATE,
    INDY_DID_EXAMPLE,
    INDY_DID_VALIDATE,
    INDY_SCHEMA_ID_EXAMPLE,
    INDY_SCHEMA_ID_VALIDATE,
    MAJOR_MINOR_VERSION_EXAMPLE,
    MAJOR_MINOR_VERSION_VALIDATE,
)

CRED_DEF_SENT_RECORD_TYPE = "cred_def_sent"


class CredDefQueryStringSchema(OpenAPISchema):
    """Query string parameters for credential definition searches."""

    schema_id = fields.Str(
        required=False,
        validate=INDY_SCHEMA_ID_VALIDATE,
        metadata={
            "description": "Schema identifier",
            "example": INDY_SCHEMA_ID_EXAMPLE,
        },
    )
    schema_issuer_did = fields.Str(
        required=False,
        validate=INDY_DID_VALIDATE,
        metadata={"description": "Schema issuer DID", "example": INDY_DID_EXAMPLE},
    )
    schema_name = fields.Str(
        required=False, metadata={"description": "Schema name", "example": "membership"}
    )
    schema_version = fields.Str(
        required=False,
        validate=MAJOR_MINOR_VERSION_VALIDATE,
        metadata={
            "description": "Schema version",
            "example": MAJOR_MINOR_VERSION_EXAMPLE,
        },
    )
    issuer_did = fields.Str(
        required=False,
        validate=INDY_DID_VALIDATE,
        metadata={"description": "Issuer DID", "example": INDY_DID_EXAMPLE},
    )
    cred_def_id = fields.Str(
        required=False,
        validate=INDY_CRED_DEF_ID_VALIDATE,
        metadata={
            "description": "Credential definition id",
            "example": INDY_CRED_DEF_ID_EXAMPLE,
        },
    )


CRED_DEF_TAGS = list(vars(CredDefQueryStringSchema).get("_declared_fields", []))

CRED_DEF_EVENT_PREFIX = "acapy::CRED_DEF::"
EVENT_LISTENER_PATTERN = re.compile(f"^{CRED_DEF_EVENT_PREFIX}(.*)?$")


async def notify_cred_def_event(profile: Profile, cred_def_id: str, meta_data: dict):
    """Send notification for a cred def post-process event."""
    await profile.notify(
        CRED_DEF_EVENT_PREFIX + cred_def_id,
        meta_data,
    )



"""Mock utilities."""

from unittest.mock import ANY, AsyncMock, MagicMock, Mock, create_autospec, patch


def CoroutineMock(*args, **kwargs):
    """Return an AsyncMock that returns a MagicMock, unless return_value is set."""
    if "return_value" in kwargs:
        return AsyncMock(*args, **kwargs)
    return AsyncMock(*args, **kwargs, return_value=MagicMock())


__all__ = [
    "CoroutineMock",
    "AsyncMock",
    "MagicMock",
    "patch",
    "create_autospec",
    "Mock",
    "ANY",
]
